import jsonpatch
import pytest
from marvin._compat import model_dump
from marvin.components.ai_application import (
    AIApplication,
    AppPlan,
    FreeformState,
    TaskState,
    UpdatePlan,
    UpdateState,
)
from marvin.tools import Tool
from marvin.utilities.messages import Message

from tests.utils.mark import pytest_mark_class


class GetSchleeb(Tool):
    name: str = "get_schleeb"

    async def run(self):
        """Get the value of schleeb"""
        return 42


class TestStateJSONPatch:
    def test_update_app_state_valid_patch(self):
        app = AIApplication(
            state=FreeformState(state={"foo": "bar"}), description="test app"
        )
        tool = UpdateState(app=app)
        tool.run([{"op": "replace", "path": "/state/foo", "value": "baz"}])
        assert model_dump(app.state) == {"state": {"foo": "baz"}}

    def test_update_app_state_invalid_patch(self):
        app = AIApplication(
            state=FreeformState(state={"foo": "bar"}), description="test app"
        )
        tool = UpdateState(app=app)
        with pytest.raises(jsonpatch.InvalidJsonPatch):
            tool.run([{"op": "invalid_op", "path": "/state/foo", "value": "baz"}])
        assert model_dump(app.state) == {"state": {"foo": "bar"}}

    def test_update_app_state_non_existent_path(self):
        app = AIApplication(
            state=FreeformState(state={"foo": "bar"}), description="test app"
        )
        tool = UpdateState(app=app)
        with pytest.raises(jsonpatch.JsonPatchConflict):
            tool.run([{"op": "replace", "path": "/state/baz", "value": "qux"}])
        assert model_dump(app.state) == {"state": {"foo": "bar"}}


@pytest_mark_class("llm")
class TestUpdateState:
    def test_keep_app_state(self):
        app = AIApplication(
            name="location tracker app",
            state=FreeformState(state={"San Francisco": {"visited": False}}),
            description="keep track of where I've visited",
        )

        app("I just visited to San Francisco")
        assert bool(app.state.state.get("San Francisco", {}).get("visited"))

        app("oh also I visited San Jose!")

        assert bool(app.state.state.get("San Jose", {}).get("visited"))

    def test_keep_app_state_undo_previous_patch(self):
        app = AIApplication(
            name="location tracker app",
            state=FreeformState(state={"San Francisco": {"visited": False}}),
            description="keep track of where I've visited",
        )

        app("I just visited San Francisco")
        assert bool(app.state.state.get("San Francisco", {}).get("visited"))

        app(
            "sorry, I was confused, I didn't visit San Francisco - but I did visit San"
            " Jose"
        )

        assert not bool(app.state.state.get("San Francisco", {}).get("visited"))
        assert bool(app.state.state.get("San Jose", {}).get("visited"))


class TestPlanJSONPatch:
    def test_update_app_plan_valid_patch(self):
        app = AIApplication(
            plan=AppPlan(
                tasks=[{"id": 1, "description": "test task", "state": "IN_PROGRESS"}]
            ),
            description="test app",
        )
        tool = UpdatePlan(app=app)
        tool.run([{"op": "replace", "path": "/tasks/0/state", "value": "COMPLETED"}])
        assert model_dump(app.plan) == {
            "tasks": [
                {
                    "id": 1,
                    "description": "test task",
                    "state": TaskState.COMPLETED,
                    "upstream_task_ids": None,
                    "parent_task_id": None,
                }
            ],
            "notes": [],
        }

    def test_update_app_plan_invalid_patch(self):
        app = AIApplication(
            plan=AppPlan(
                tasks=[{"id": 1, "description": "test task", "state": "IN_PROGRESS"}]
            ),
            description="test app",
        )
        tool = UpdatePlan(app=app)
        with pytest.raises(jsonpatch.JsonPatchException):
            tool.run(
                [{"op": "invalid_op", "path": "/tasks/0/state", "value": "COMPLETED"}]
            )
        assert model_dump(app.plan) == {
            "tasks": [
                {
                    "id": 1,
                    "description": "test task",
                    "state": TaskState.IN_PROGRESS,
                    "upstream_task_ids": None,
                    "parent_task_id": None,
                }
            ],
            "notes": [],
        }

    def test_update_app_plan_non_existent_path(self):
        app = AIApplication(
            plan=AppPlan(
                tasks=[{"id": 1, "description": "test task", "state": "IN_PROGRESS"}]
            ),
            description="test app",
        )
        tool = UpdatePlan(app=app)
        with pytest.raises(jsonpatch.JsonPointerException):
            tool.run(
                [{"op": "replace", "path": "/tasks/1/state", "value": "COMPLETED"}]
            )
        assert model_dump(app.plan) == {
            "tasks": [
                {
                    "id": 1,
                    "description": "test task",
                    "state": TaskState.IN_PROGRESS,
                    "upstream_task_ids": None,
                    "parent_task_id": None,
                }
            ],
            "notes": [],
        }


@pytest_mark_class("llm")
class TestUpdatePlan:
    def test_keep_app_plan(self):
        app = AIApplication(
            name="Zoo planner app",
            plan=AppPlan(
                tasks=[
                    {
                        "id": 1,
                        "description": "Visit tigers",
                        "state": TaskState.IN_PROGRESS,
                    },
                    {
                        "id": 2,
                        "description": "Visit giraffes",
                        "state": TaskState.PENDING,
                    },
                ]
            ),
            description="plan and track my visit to the zoo",
        )

        app(
            "Actually I heard the tigers ate Carol Baskin's husband - I think I'll skip"
            " visiting them."
        )

        assert [task["state"] for task in app.plan.dict()["tasks"]] == [
            TaskState.SKIPPED,
            TaskState.PENDING,
        ]

        app("Dude i just visited the giraffes!")

        assert [task["state"] for task in app.plan.dict()["tasks"]] == [
            TaskState.SKIPPED,
            TaskState.COMPLETED,
        ]


@pytest_mark_class("llm")
class TestUseCallable:
    def test_use_sync_fn(self):
        def get_schleeb():
            return 42

        app = AIApplication(
            name="Schleeb app",
            tools=[get_schleeb],
            state_enabled=False,
            plan_enabled=False,
            description="answer user questions",
        )

        assert "42" in app("what is the value of schleeb?").content

    def test_use_async_fn(self):
        async def get_schleeb():
            return 42

        app = AIApplication(
            name="Schleeb app",
            tools=[get_schleeb],
            state_enabled=False,
            plan_enabled=False,
            description="answer user questions",
        )

        assert "42" in app("what is the value of schleeb?").content


@pytest_mark_class("llm")
class TestUseTool:
    def test_use_tool(self):
        app = AIApplication(
            name="Schleeb app",
            tools=[GetSchleeb()],
            state_enabled=False,
            plan_enabled=False,
            description="answer user questions",
        )

        assert "42" in app("what is the value of schleeb?").content


@pytest_mark_class("llm")
class TestStreaming:
    def test_streaming(self):
        external_state = {"content": []}

        app = AIApplication(
            name="streaming app",
            stream_handler=lambda m: external_state["content"].append(m.content),
            state_enabled=False,
            plan_enabled=False,
        )

        response = app(
            "say the words 'Hello world' EXACTLY as i have written them."
            " no other characters should be included, do not add any punctuation."
        )

        assert isinstance(response, Message)
        assert response.content == "Hello world"

        assert external_state["content"] == ["", "Hello", "Hello world", "Hello world"]



import sys

if sys.platform == "win32" and sys.version_info > (3, 10):
    import collections.abc

    # Create a new module to replace the collections module
    class ShimModule:
        def __init__(self, original_module):
            self.__original_module = original_module

        def __getattr__(self, name):
            if name == "Callable":
                return collections.abc.Callable
            return getattr(self.__original_module, name)

    # Replace the collections module in sys.modules with the shim
    sys.modules["collections"] = ShimModule(sys.modules["collections"])



import sys
from importlib import import_module


def cached_import(module_path, class_name):
    """
    This function checks if the module is already imported and if it is,
    it returns the class. Otherwise, it imports the module and returns the class.
    """

    # Check if the module is already imported

    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, "__spec__", None))
        and getattr(spec, "_initializing", False) is False
    ):
        module = import_module(module_path)
    return getattr(module, class_name)


def import_string(path: str):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        path_, class_ = path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % path) from err
    try:
        return cached_import(path_, class_)
    except AttributeError as err:
        raise ImportError(f"Module '{path_}' isn't a '{class_}' attr/class") from err



import inspect
from typing import Any, AsyncGenerator, Callable, Optional, TypeVar, Union

from marvin._compat import cast_to_json, model_dump
from marvin.settings import settings
from marvin.types import Function
from marvin.utilities.async_utils import create_task
from marvin.utilities.messages import Message
from marvin.utilities.streaming import StreamHandler
from openai.openai_object import OpenAIObject
from pydantic import BaseModel

from ..abstract import AbstractChatCompletion
from ..handlers import Request, Response, Usage

T = TypeVar(
    "T",
    bound=BaseModel,
)

CONTEXT_SIZES = {
    "gpt-3.5-turbo-16k-0613": 16384,
    "gpt-3.5-turbo-16k": 16384,
    "gpt-3.5-turbo-0613": 4096,
    "gpt-3.5-turbo": 4096,
    "gpt-4-32k-0613": 32768,
    "gpt-4-32k": 32768,
    "gpt-4-0613": 8192,
    "gpt-4": 8192,
}


def serialize_function_or_callable(
    function_or_callable: Union[Function, Callable[..., Any]],
    name: Optional[str] = None,
    description: Optional[str] = None,
    field_name: Optional[str] = None,
) -> dict[str, Any]:
    if isinstance(function_or_callable, Function):
        return {
            "name": function_or_callable.__name__,
            "description": function_or_callable.__doc__,
            "parameters": function_or_callable.schema,
        }
    else:
        return cast_to_json(
            function_or_callable,
            name=name,
            description=description,
            field_name=field_name,
        )


class OpenAIStreamHandler(StreamHandler):
    async def handle_streaming_response(
        self,
        api_response: AsyncGenerator[OpenAIObject, None],
    ) -> OpenAIObject:
        final_chunk = {}
        accumulated_content = ""

        async for r in api_response:
            final_chunk.update(r.to_dict_recursive())

            delta = r.choices[0].delta if r.choices and r.choices[0] else None

            if delta is None:
                continue

            if "content" in delta:
                accumulated_content += delta.content or ""

            if self.callback:
                callback_result = self.callback(
                    Message(
                        content=accumulated_content, role="assistant", data=final_chunk
                    )
                )
                if inspect.isawaitable(callback_result):
                    create_task(callback_result)

        if "choices" in final_chunk and len(final_chunk["choices"]) > 0:
            final_chunk["choices"][0]["content"] = accumulated_content

        final_chunk["object"] = "chat.completion"

        return OpenAIObject.construct_from(
            {
                "id": final_chunk["id"],
                "object": "chat.completion",
                "created": final_chunk["created"],
                "model": final_chunk["model"],
                "choices": [
                    {
                        "index": 0,
                        "message": {
                            "role": "assistant",
                            "content": accumulated_content,
                        },
                        "finish_reason": "stop",
                    }
                ],
                # TODO: Figure out how to get the usage from the streaming response
                "usage": Usage.parse_obj(
                    {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0}
                ),
            }
        )


class OpenAIChatCompletion(AbstractChatCompletion[T]):
    """
    OpenAI-specific implementation of the ChatCompletion.
    """

    def __init__(self, provider: str, **kwargs: Any):
        super().__init__(defaults=settings.get_defaults(provider or "openai") | kwargs)

    def _serialize_request(
        self, request: Optional[Request[T]] = None
    ) -> dict[str, Any]:
        """
        Serialize the request as per OpenAI's requirements.
        """
        _request = request or Request()
        _request = Request(
            **self.defaults
            | (
                model_dump(
                    request,
                    exclude_none=True,
                )
                if request
                else {}
            )
        )
        _request.function_call = _request.function_call or (
            request and request.function_call
        )
        _request.functions = _request.functions or (request and request.functions)
        _request.response_model = _request.response_model or (
            request and request.response_model
        )  # noqa

        extras = model_dump(
            _request,
            exclude={"functions", "function_call", "response_model"},
        )

        functions: dict[str, Any] = {}
        function_call: Any = {}
        for message in extras.get("messages", []):
            if message.get("name", -1) is None:
                message.pop("name", None)
            if message.get("function_call", -1) is None:
                message.pop("function_call", None)

        if _request.response_model:
            schema = cast_to_json(_request.response_model)
            functions["functions"] = [schema]
            function_call["function_call"] = {"name": schema.get("name")}

        elif _request.functions:
            functions["functions"] = [
                serialize_function_or_callable(function)
                for function in _request.functions
            ]
            if _request.function_call:
                function_call["function_call"] = _request.function_call
        return extras | functions | function_call

    def _create_request(self, **kwargs: Any) -> Request[T]:
        """
        Prepare and return an OpenAI-specific request object.
        """
        return Request(**kwargs)

    def _parse_response(self, response: Any) -> Response[T]:
        """
        Parse the response received from OpenAI.
        """
        # Convert OpenAI's response into a standard format or object
        return Response(**response.to_dict_recursive())  # type: ignore

    def _send_request(self, **serialized_request: Any) -> Any:
        """
        Send the serialized request to OpenAI's endpoint/service.
        """
        # Use openai's library functions to send the request and get a response
        # Example:

        import openai

        response = openai.ChatCompletion.create(**serialized_request)  # type: ignore
        return response  # type: ignore

    async def _send_request_async(self, **serialized_request: Any) -> Response[T]:
        """
        Send the serialized request to OpenAI's endpoint asynchronously.
        """
        import openai

        if handler_fn := serialized_request.pop("stream_handler", {}):
            serialized_request["stream"] = True

        response = await openai.ChatCompletion.acreate(**serialized_request)  # type: ignore # noqa

        if handler_fn:
            response = await OpenAIStreamHandler(
                callback=handler_fn,
            ).handle_streaming_response(response)

        return response  # type: ignore



import warnings
from types import GenericAlias
from typing import Any, Union

import pydantic
from pydantic import BaseModel, Field, PrivateAttr

import marvin
import marvin.utilities.types
from marvin.tools import Tool
from marvin.utilities.types import (
    genericalias_contains,
    safe_issubclass,
)

SENTINEL = "__SENTINEL__"


class FormatResponse(Tool):
    _cached_type: Union[type, GenericAlias] = PrivateAttr(SENTINEL)
    type_schema: dict[str, Any] = Field(
        ..., description="The OpenAPI schema for the type"
    )
    description: str = (
        "You MUST always call this function before responding to the user to ensure"
        " that your final response is formatted correctly and complies with the output"
        " format requirements."
    )

    def __init__(self, type_: Union[type, GenericAlias] = SENTINEL, **kwargs):
        if type_ is not SENTINEL:
            if not isinstance(type_, (type, GenericAlias)):
                raise ValueError(f"Expected a type or GenericAlias, got {type_}")

            # warn if the type is a set or tuple with GPT 3.5
            if (
                "gpt-3.5" in marvin.settings.llm_model
                or "gpt-35" in marvin.settings.llm_model
            ):
                if safe_issubclass(type_, (set, tuple)) or genericalias_contains(
                    type_, (set, tuple)
                ):
                    warnings.warn(
                        (
                            "GPT-3.5 often fails with `set` or `tuple` types. Consider"
                            " using `list` instead."
                        ),
                        UserWarning,
                    )

            type_schema = marvin.utilities.types.type_to_schema(
                type_, set_root_type=False
            )
            type_schema.pop("title", None)
            kwargs["type_schema"] = type_schema

        super().__init__(**kwargs)
        if type_ is not SENTINEL:
            if type_schema.get("description"):
                self.description += f"\n\n {type_schema['description']}"

        if type_ is not SENTINEL:
            self._cached_type = type_

    def get_type(self) -> Union[type, GenericAlias]:
        if self._cached_type is not SENTINEL:
            return self._cached_type
        model = marvin.utilities.types.schema_to_type(self.type_schema)
        type_ = model.__fields__["__root__"].outer_type_
        self._cached_type = type_
        return type_

    def run(self, **kwargs) -> Any:
        type_ = self.get_type()
        if not safe_issubclass(type_, BaseModel):
            kwargs = kwargs["data"]
        return pydantic.parse_obj_as(type_, kwargs)

    def argument_schema(self) -> dict:
        return self.type_schema



from datetime import datetime

from marvin import AIApplication
from pydantic import BaseModel, Field


class ToDo(BaseModel):
    title: str
    description: str = None
    due_date: datetime = None
    done: bool = False


class ToDoState(BaseModel):
    todos: list[ToDo] = []


class ToDoApp(AIApplication):
    state: ToDoState = Field(default_factory=ToDoState)
    description: str = """
        A simple to-do tracker. Users will give instructions to add, remove, and
        update their to-dos.
        """
    plan_enabled: bool = False


__all__ = ["ToDoApp"]



from typing import Any, Callable, Dict, List, Optional, Set, Type, Union
from fastapi.routing import APIRouter
from pydantic import BaseModel, validate_arguments
from marvin.utilities.types import function_to_model
from marvin.utilities.messages import Message
from marvin.openai.Function import openai_fn
from openai.openai_object import OpenAIObject
from marvin.functions import FunctionRegistry


class OpenAIFunctionRegistry(FunctionRegistry):
    def __init__(self, function_decorator=openai_fn, *args, **kwargs):
        self.function_decorator = function_decorator
        super().__init__(function_decorator=function_decorator, *args, **kwargs)

    def from_response(self, response: OpenAIObject) -> Any:
        return next(
            iter(
                [
                    {"name": k, "content": v}
                    for k, v in self.dict_from_openai_response(response).items()
                    if v is not None
                ]
            ),
            None,
        )

    def dict_from_openai_response(self, response: OpenAIObject) -> Any:
        return {
            fn.name: fn.from_response(response)
            for fn in map(lambda fn: self.function_decorator(fn=fn), self.endpoints)
        }



import abc
import inspect
from functools import partial, wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
)

from jinja2 import Environment
from pydantic import BaseModel, Field
from typing_extensions import ParamSpec, Self

import marvin
from marvin._compat import cast_to_json, cast_to_model, model_dump, model_json_schema
from marvin.core.ChatCompletion import ChatCompletion
from marvin.core.ChatCompletion.abstract import AbstractChatCompletion
from marvin.utilities.messages import Message, Role
from marvin.utilities.strings import count_tokens, jinja_env

T = TypeVar("T")
P = ParamSpec("P")


class MessageList(list[Message]):
    def render(
        self: Self,
        **kwargs: Any,
    ) -> Self:
        return render_prompts(self, render_kwargs=kwargs)

    def serialize(
        self: Self,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        return [model_dump(message) for message in self.render(**kwargs)]


class PromptList(list[Union["Prompt", Message]]):
    def __init__(self, prompts: list[Union["Prompt", Message]]):
        super().__init__(prompts)

    def render(
        self: Self,
        content: Optional[str] = None,
        render_kwargs: Optional[dict[str, Any]] = None,
    ) -> list[Message]:
        return render_prompts(self, render_kwargs=render_kwargs)

    def dict(self, **kwargs: Any):
        return [model_dump(message) for message in self.render(**kwargs)]

    def __call__(self, **kwargs: Any):
        return self.render(**kwargs)


class BasePrompt(BaseModel, abc.ABC):
    """
    Base class for prompt templates.
    """

    functions: Optional[
        Union[
            List[Union[Dict[str, Any], Callable[..., Any], type[BaseModel]]],
            Callable[
                ..., List[Union[Dict[str, Any], Callable[..., Any], type[BaseModel]]]
            ],
        ]
    ] = Field(default=None)

    function_call: Optional[
        Union[
            Literal["auto"],
            Dict[Literal["name"], str],
        ]
    ] = Field(default=None)

    response_model: Optional[
        Union[
            type,
            type[BaseModel],
            Any,
            Callable[..., Union[type, type[BaseModel], Any]],
        ]
    ] = Field(default=None)

    response_model_name: Optional[str] = Field(
        default=None,
        exclude=True,
        repr=False,
    )
    response_model_description: Optional[str] = Field(
        default=None,
        exclude=True,
        repr=False,
    )
    response_model_field_name: Optional[str] = Field(
        default=None,
        exclude=True,
        repr=False,
    )

    position: Optional[int] = Field(
        default=None,
        repr=False,
        exclude=True,
        description=(
            "Position indicates the desired index for this prompt's messages. 0"
            " indicates they should be first; 1 indicates they should be second; -1"
            " indicates they should be last; None indicates they should be between any"
            " prompts that do request a position."
        ),
    )
    priority: float = Field(
        default=10,
        repr=False,
        exclude=True,
        description=(
            "Priority indicates the weight given when trimming messages to satisfy"
            " context limitations. Lower numbers indicate higher priority e.g. the"
            " highest priority is 0. The default is 10. Ties will be broken by message"
            " timestamp and role."
        ),
    )

    @abc.abstractmethod
    def generate(self, **kwargs: Any) -> list["Message"]:
        """
        Abstract method that generates a list of messages from the prompt template
        """
        pass

    def render(
        self: Self, content: str, render_kwargs: Optional[dict[str, Any]] = None
    ) -> str:
        """
        Helper function for rendering any jinja2 template with runtime render kwargs
        """
        return jinja_env.from_string(inspect.cleandoc(content)).render(
            **(render_kwargs or {})
        )

    def __or__(self: Self, other: Union[Self, list[Self]]) -> PromptList:
        """
        Supports pipe syntax:
        prompt = (
            Prompt()
            | Prompt()
            | Prompt()
        )
        """
        # when the right operand is a Prompt object
        if isinstance(other, Prompt):
            return PromptList([self, other])
        # when the right operand is a list
        elif isinstance(other, list[Prompt]):
            return PromptList([self, *other])
        else:
            raise TypeError(
                f"unsupported operand type(s) for |: '{type(self).__name__}' and"
                f" '{type(other).__name__}'"
            )

    def __ror__(self, other):
        """
        Supports pipe syntax:
        prompt = (
            Prompt()
            | Prompt()
            | Prompt()
        )
        """
        # when the left operand is a Prompt object
        if isinstance(other, Prompt):
            return PromptList([other, self])
        # when the left operand is a list
        elif isinstance(other, list):
            return PromptList(other + [self])
        else:
            raise TypeError(
                f"unsupported operand type(s) for |: '{type(other).__name__}' and"
                f" '{type(self).__name__}'"
            )


class Prompt(BasePrompt, Generic[P], extra="allow", arbitrary_types_allowed=True):
    def generate(self, **kwargs: Any) -> list[Message]:
        response = Message.from_transcript(
            self.render(content=self.__doc__ or "", render_kwargs=kwargs)
        )
        return response

    def to_dict(self, **kwargs: Any) -> dict[str, Any]:
        extras = model_dump(
            self,
            exclude=set(self.__fields__.keys()),
            exclude_none=True,
        )
        return {
            "messages": [
                model_dump(message, include={"content", "role"})
                for message in render_prompts(
                    self.generate(
                        **extras | kwargs | {"response_model": self.response_model}
                    )
                )
            ],
            "functions": self.functions,
            "function_call": self.function_call,
            "response_model": self.response_model,
        }

    def to_chat_completion(
        self, model: Optional[str] = None, **model_kwargs: Any
    ) -> AbstractChatCompletion[T]:
        return ChatCompletion(model=model, **model_kwargs)(**self.to_dict())

    def serialize(self, model: Any = None, **kwargs: Any) -> dict[str, Any]:
        if model:
            return model(**self.to_dict(**kwargs))._serialize_request()  # type: ignore

        _dict = self.to_dict(**kwargs)

        response: dict[str, Any] = {}
        response["messages"] = _dict["messages"]

        if _dict.get("response_model", None):
            response["functions"] = [
                model_json_schema(cast_to_model(_dict["response_model"]))
            ]
            response["function_call"] = {"name": response["functions"][0]["name"]}
        elif _dict.get("functions", None):
            response["functions"] = [
                cast_to_json(function) if callable(function) else function
                for function in _dict["functions"]
            ]
            if _dict["function_call"]:
                response["function_call"] = _dict["function_call"]

        return response

    @classmethod
    def as_decorator(
        cls: type[Self],
        func: Optional[Callable[P, Any]] = None,
        *,
        environment: Optional[Environment] = None,
        ctx: Optional[dict[str, Any]] = None,
        role: Optional[Role] = None,
        functions: Optional[
            list[Union[Callable[..., Any], type[BaseModel], dict[str, Any]]]
        ] = None,  # noqa
        function_call: Optional[
            Union[Literal["auto"], dict[Literal["name"], str]]
        ] = None,  # noqa
        response_model: Optional[type[BaseModel]] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        serialize_on_call: bool = True,
    ) -> Union[
        Callable[[Callable[P, None]], Callable[P, None]],
        Callable[[Callable[P, None]], Callable[P, Self]],
        Callable[P, Self],
    ]:
        def wrapper(func: Callable[P, Any], *args: P.args, **kwargs: P.kwargs) -> Self:
            signature = inspect.signature(func)
            params = signature.bind(*args, **kwargs)
            params.apply_defaults()
            response = type(getattr(cls, "__name__", ""), (cls,), {})(
                __params__=params.arguments,
                **params.arguments,
                **ctx or {},
                functions=functions,
                function_call=function_call,
                response_model=cast_to_model(
                    response_model or signature.return_annotation,
                    name=response_model_name,
                    description=response_model_description,
                    field_name=response_model_field_name,
                ),
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
            )
            response.__doc__ = func.__doc__
            if serialize_on_call:
                return response.serialize()
            return response

        if func is not None:
            return wraps(func)(partial(wrapper, func))

        def decorator(func: Callable[P, None]) -> Callable[P, Self]:
            return wraps(func)(partial(wrapper, func))

        return decorator


prompt_fn = Prompt.as_decorator


class MessageWrapper(BasePrompt):
    """
    A Prompt class that stores and returns a specific Message
    """

    message: Message

    def generate(self, **kwargs: Any) -> list[Message]:
        return [self.message]


def render_prompts(
    prompts: Union[list[Message], list[Union[Prompt, Message]]],
    render_kwargs: Optional[dict[str, Any]] = None,
    max_tokens: Optional[int] = None,
) -> MessageList:
    max_tokens = max_tokens or marvin.settings.llm_max_context_tokens

    all_messages = []

    # if the user supplied any messages, wrap them in a MessageWrapper so we can
    # treat them as prompts for sorting and filtering
    prompts = [
        MessageWrapper(message=p) if isinstance(p, Message) else p for p in prompts
    ]

    # Separate prompts by positive, none and negative position
    pos_prompts = [p for p in prompts if p.position is not None and p.position >= 0]
    none_prompts = [p for p in prompts if p.position is None]
    neg_prompts = [p for p in prompts if p.position is not None and p.position < 0]

    # Sort the positive prompts in ascending order and negative prompts in
    # descending order, but both with timestamp ascending
    pos_prompts = sorted(pos_prompts, key=lambda c: c.position)
    neg_prompts = sorted(neg_prompts, key=lambda c: c.position, reverse=True)

    # generate messages from all prompts
    for i, prompt in enumerate(pos_prompts + none_prompts + neg_prompts):
        prompt_messages = prompt.generate(**(render_kwargs or {})) or []
        all_messages.extend((prompt.priority, i, m) for m in prompt_messages)

    # sort all messages by (priority asc, position desc)  and stop when the
    # token limit is reached. This will prefer high-priority messages that are
    # later in the message chain.
    current_tokens = 0
    allowed_messages = []
    for _, position, msg in sorted(all_messages, key=lambda m: (m[0], -1 * m[1])):
        if current_tokens >= max_tokens or not (content := msg.content):
            break
        allowed_messages.append((position, msg))
        current_tokens += count_tokens(content)

    # sort allowed messages by position to restore original order
    messages = [msg for _, msg in sorted(allowed_messages, key=lambda m: m[0])]

    # Combine all system messages into one and insert at the index of the first
    # system message
    system_messages = [m for m in messages if m.role == Role.SYSTEM]
    if len(system_messages) > 1:
        system_message = Message(
            role=Role.SYSTEM,
            content="\n\n".join([m.content for m in system_messages]),
        )
        system_message_index = messages.index(system_messages[0])
        messages = [m for m in messages if m.role != Role.SYSTEM]
        messages.insert(system_message_index, system_message)

    # return all messages
    return messages



from typing import Optional

from pydantic import BaseModel

from marvin.components.ai_model import ai_model


class DataSchema(BaseModel):
    title: Optional[str] = None
    type: Optional[str] = None
    properties: Optional[dict] = {}
    required: Optional[list[str]] = []
    additionalProperties: bool = False
    definitions: dict = {}
    description: Optional[str] = None


# If you're reading this and expected something fancier,
# I'm sorry to disappoint you. It's this simple.
AIModelFactory = ai_model(DataSchema)



import asyncio
import logging
import sys

import pytest

from .fixtures import *


@pytest.fixture(scope="session")
def event_loop(request):
    """
    Redefine the event loop to support session/module-scoped fixtures;
    see https://github.com/pytest-dev/pytest-asyncio/issues/68

    When running on Windows we need to use a non-default loop for subprocess support.
    """
    if sys.platform == "win32" and sys.version_info >= (3, 8):
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())

    policy = asyncio.get_event_loop_policy()

    loop: asyncio.BaseEventLoop = policy.new_event_loop()

    # configure asyncio logging to capture long running tasks
    asyncio_logger = logging.getLogger("asyncio")
    asyncio_logger.setLevel("WARNING")
    asyncio_logger.addHandler(logging.StreamHandler())
    loop.set_debug(True)
    loop.slow_callback_duration = 0.25

    try:
        yield loop
    finally:
        # ensure all background tasks are cancelled and awaited to avoid
        # spurious warnings / errors about "task destroyed but it is pending"
        tasks = asyncio.all_tasks(loop=loop)
        for task in tasks:
            task.cancel()
        if tasks and loop.is_running():
            loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
            loop.close()



import json
from typing import TypeVar
from openai.openai_object import OpenAIObject

from pydantic import validate_arguments
from marvin.functions import Function, FunctionDecoratorFactory


T = TypeVar("T")
A = TypeVar("A")


class OpenAIFunction(Function):
    def __call__(self, response: OpenAIObject) -> T:
        return self.from_response(response)

    @validate_arguments
    def from_response(self, response: OpenAIObject) -> T:
        relevant_calls = [
            choice.message.function_call
            for choice in response.choices
            if hasattr(choice.message, "function_call")
            and self.name == choice.message.function_call.get("name", None)
        ]

        arguments = [
            json.loads(function_call.get("arguments"))
            for function_call in relevant_calls
        ]

        responses = [self.fn(**argument) for argument in arguments]

        if len(responses) == 0:
            return None
        elif len(responses) == 1:
            return responses[0]
        else:
            return responses


openai_fn = FunctionDecoratorFactory(name="openai", func_class=OpenAIFunction)



import inspect
import uuid
from datetime import datetime
from enum import Enum
from typing import Any, Optional
from zoneinfo import ZoneInfo

from pydantic import BaseModel, Field
from typing_extensions import Self

from marvin._compat import field_validator
from marvin.utilities.strings import split_text_by_tokens
from marvin.utilities.types import MarvinBaseModel


class Role(Enum):
    SYSTEM = "system"
    ASSISTANT = "assistant"
    USER = "user"
    FUNCTION_REQUEST = "function"
    FUNCTION_RESPONSE = "function"

    @classmethod
    def _missing_(cls: type[Self], value: object) -> Optional[Self]:
        lower_value = str(value).lower()
        matching_member = next(
            (member for member in cls if member.value.lower() == lower_value), None
        )
        return matching_member


class FunctionCall(BaseModel):
    name: str
    arguments: str


class Message(MarvinBaseModel):
    role: Role
    content: Optional[str] = Field(default=None, description="The message content")

    name: Optional[str] = Field(
        default=None,
        description="The name of the message",
    )

    function_call: Optional[FunctionCall] = Field(default=None)

    # convenience fields, excluded from serialization
    id: uuid.UUID = Field(default_factory=uuid.uuid4, exclude=True)
    data: Optional[dict[str, Any]] = Field(default_factory=dict, exclude=True)
    timestamp: datetime = Field(
        default_factory=lambda: datetime.now(ZoneInfo("UTC")),
        exclude=True,
    )

    @field_validator("content")
    def clean_content(cls, v: Optional[str]) -> Optional[str]:
        if v is not None:
            v = inspect.cleandoc(v)
        return v

    class Config(MarvinBaseModel.Config):
        use_enum_values = True

    @classmethod
    def from_transcript(
        cls: type[Self],
        text: str,
    ) -> list[Self]:
        pairs = split_text_by_tokens(
            text=text, split_tokens=[role.value.capitalize() for role in Role]
        )
        return [
            cls(
                role=Role(pair[0]),
                content=pair[1],
            )
            for pair in pairs
        ]



import itertools
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, TypeVar

T = TypeVar("T")


def batched(
    iterable: Iterable[T], size: int, size_fn: Callable[[Any], int] = None
) -> Iterable[T]:
    """
    If size_fn is not provided, then the batch size will be determined by the
    number of items in the batch.

    If size_fn is provided, then it will be used
    to compute the batch size. Note that if a single item is larger than the
    batch size, it will be returned as a batch of its own.

    Args:
        iterable: The iterable to batch.
        size: The size of each batch.
        size_fn: A function that takes an item from the iterable and returns its size.

    Returns:
        An iterable of batches.

    Example:
        Batch a list of integers into batches of size 2:
        ```python
        batched([1, 2, 3, 4, 5], 2)
        # [[1, 2], [3, 4], [5]]
        ```
    """
    if size_fn is None:
        it = iter(iterable)
        while True:
            batch = tuple(itertools.islice(it, size))
            if not batch:
                break
            yield batch
    else:
        batch = []
        batch_size = 0
        for item in iter(iterable):
            batch.append(item)
            batch_size += size_fn(item)
            if batch_size > size:
                yield batch
                batch = []
                batch_size = 0
        if batch:
            yield batch


def multi_glob(
    directory: Optional[str] = None,
    keep_globs: Optional[list[str]] = None,
    drop_globs: Optional[list[str]] = None,
) -> list[Path]:
    """Return a list of files in a directory that match the given globs.

    Args:
        directory: The directory to search. Defaults to the current working directory.
        keep_globs: A list of globs to keep. Defaults to ["**/*"].
        drop_globs: A list of globs to drop. Defaults to [".git/**/*"].

    Returns:
        A list of `Path` objects in the directory that match the given globs.

    Example:
        Recursively find all Python files in the `src` directory:
        ```python
        all_python_files = multi_glob(directory="src", keep_globs=["**/*.py"])
        ```
    """
    keep_globs = keep_globs or ["**/*"]
    drop_globs = drop_globs or [".git/**/*"]

    directory_path = Path(directory) if directory else Path.cwd()

    if not directory_path.is_dir():
        raise ValueError(f"'{directory}' is not a directory.")

    def files_from_globs(globs):
        return {
            file
            for pattern in globs
            for file in directory_path.glob(pattern)
            if file.is_file()
        }

    matching_files = files_from_globs(keep_globs) - files_from_globs(drop_globs)

    return [file.relative_to(directory_path) for file in matching_files]



import typer
import os
import shutil
from jinja2 import Template

from pathlib import Path
from marvin.cli.admin.scripts.create_env_file import create_env_file
from marvin.cli.admin.scripts.create_secure_key import create_secure_key
from marvin._framework._defaults import default_settings

# Get the absolute path of the current file
filename = Path(__file__).resolve()

# Navigate two directories back from the current file
source_path = filename.parent.parent.parent / "_framework"

app = typer.Typer()


@app.command()
def startproject(no_input: bool = False):
    project_name = typer.prompt("Project Name")
    openai_api_key = typer.prompt("OpenAI API Key")
    shutil.copytree(
        source_path,
        os.path.join(os.getcwd(), project_name),
        ignore=shutil.ignore_patterns(
            "@*"
        ),  # This will ignore all files/directories starting with @
    )
    with open(
        os.path.join(os.getcwd(), project_name, "config/settings.py"), "r"
    ) as file:
        template = Template(file.read())
        rendered = template.render(
            **default_settings, project_name=project_name, openai_api_key=openai_api_key
        )
    with open(
        os.path.join(os.getcwd(), project_name, "config/settings.py"), "w"
    ) as rendered_file:
        rendered_file.write(rendered)
    create_env_file(
        os.path.join(os.getcwd(), project_name),
        [("MARVIN_SECRET", create_secure_key()), ("OPENAI_API_KEY", openai_api_key)],
    )


@app.command()
def startapp(no_input: bool = False):
    print("beep")


if __name__ == "__main__":
    app()



import pytest


def pytest_mark_class(marker):
    def decorator(cls):
        for attr_name, attr_value in cls.__dict__.items():
            if callable(attr_value) and attr_name.startswith("test"):
                setattr(cls, attr_name, pytest.mark.llm(attr_value))
        return cls

    return decorator



import asyncio
import inspect
from functools import partial
from typing import Any, Awaitable, Callable, Generic, Optional, TypeVar, Union

from pydantic import BaseModel, Field
from typing_extensions import ParamSpec, Self

from marvin.core.ChatCompletion import ChatCompletion
from marvin.core.ChatCompletion.abstract import AbstractChatCompletion
from marvin.prompts import Prompt, prompt_fn
from marvin.utilities.async_utils import run_sync

T = TypeVar("T", bound=BaseModel)

A = TypeVar("A", bound=Any)

P = ParamSpec("P")


def ai_fn_prompt(
    func: Callable[P, Any],
    ctx: Optional[dict[str, Any]] = None,
    **kwargs: Any,
) -> Callable[P, Prompt[P]]:
    return_annotation: Any = inspect.signature(func).return_annotation

    @prompt_fn(
        ctx={"ctx": ctx or {}, "func": func, "inspect": inspect},
        response_model=return_annotation,
        serialize_on_call=False,
        **kwargs,
    )
    def prompt_wrapper(*args: P.args, **kwargs: P.kwargs) -> None:  # type: ignore # noqa
        """
        System: {{ctx.get('instructions') if ctx.get('instructions')}}

        Your job is to generate likely outputs for a Python function with the
        following signature and docstring:

        {{'def' + ''.join(inspect.getsource(func).split('def')[1:])}}

        The user will provide function inputs (if any) and you must respond with
        the most likely result.

        User: The function was called with the following inputs:
        {% set sig = inspect.signature(func) %}
        {% set binds = sig.bind(*args, **kwargs) %}
        {% set defaults = binds.apply_defaults() %}
        {% set params = binds.arguments %}
        {%for (arg, value) in params.items()%}
        - {{ arg }}: {{ value }}
        {% endfor %}

        What is its output?
        """

    return prompt_wrapper  # type: ignore


class AIFunction(BaseModel, Generic[P, T]):
    fn: Callable[P, Any]
    ctx: Optional[dict[str, Any]] = None
    model: Any = Field(default_factory=ChatCompletion)
    response_model_name: Optional[str] = Field(default=None, exclude=True)
    response_model_description: Optional[str] = Field(default=None, exclude=True)
    response_model_field_name: Optional[str] = Field(default=None, exclude=True)

    def __call__(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Any:
        return self.call(*args, **kwargs)

    def get_prompt(
        self,
    ) -> Callable[P, Prompt[P]]:
        return ai_fn_prompt(
            self.fn,
            ctx=self.ctx,
            response_model_name=self.response_model_name,
            response_model_description=self.response_model_description,
            response_model_field_name=self.response_model_field_name,
        )

    def as_prompt(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> dict[str, Any]:
        return self.get_prompt()(*args, **kwargs).serialize(
            model=self.model,
        )

    def as_dict(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> dict[str, Any]:
        return self.get_prompt()(*args, **kwargs).to_dict()

    def as_chat_completion(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AbstractChatCompletion[T]:
        return self.model(**self.as_dict(*args, **kwargs))

    def call(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Any:
        return getattr(
            self.as_chat_completion(*args, **kwargs).create().to_model(),
            self.response_model_field_name or "output",
        )

    async def acall(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Any:
        return getattr(
            (await self.as_chat_completion(*args, **kwargs).acreate()).to_model(),
            self.response_model_field_name or "output",
        )

    def map(self, *map_args: list[Any], **map_kwargs: list[Any]):
        """
        Map the AI function over a sequence of arguments. Runs concurrently.

        Arguments should be provided as if calling the function normally, but
        each argument must be a list. The function is called once for each item
        in the list, and the results are returned in a list.

        This method should be called synchronously.

        For example, fn.map([1, 2]) is equivalent to [fn(1), fn(2)].

        fn.map([1, 2], x=['a', 'b']) is equivalent to [fn(1, x='a'), fn(2, x='b')].
        """
        return run_sync(self.amap(*map_args, **map_kwargs))

    async def amap(self, *map_args: list[Any], **map_kwargs: list[Any]) -> list[Any]:
        tasks: list[Any] = []
        if map_args:
            max_length = max(len(arg) for arg in map_args)
        else:
            max_length = max(len(v) for v in map_kwargs.values())

        for i in range(max_length):
            call_args = [arg[i] if i < len(arg) else None for arg in map_args]
            call_kwargs = (
                {k: v[i] if i < len(v) else None for k, v in map_kwargs.items()}
                if map_kwargs
                else {}
            )
            tasks.append(self.acall(*call_args, **call_kwargs))

        return await asyncio.gather(*tasks)

    @classmethod
    def as_decorator(
        cls: type[Self],
        fn: Optional[Callable[P, T]] = None,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> Union[Callable[P, T], Callable[P, Awaitable[T]]]:
        if not fn:
            return partial(
                cls.as_decorator,
                ctx=ctx,
                instructions=instructions,
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
                model=model,
                **model_kwargs,
            )  # type: ignore

        if not inspect.iscoroutinefunction(fn):
            return cls(
                fn=fn,
                ctx={"instructions": instructions, **(ctx or {})},
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
                model=ChatCompletion(model=model, **model_kwargs),
            )
        else:
            return AsyncAIFunction[P, T](
                fn=fn,
                ctx={"instructions": instructions, **(ctx or {})},
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
                model=ChatCompletion(model=model, **model_kwargs),
            )


class AsyncAIFunction(AIFunction[P, T]):
    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Any:
        return await super().acall(*args, **kwargs)


ai_fn = AIFunction.as_decorator



import inspect
from enum import Enum
from typing import Any, Callable, Optional, Union

from jsonpatch import JsonPatch
from pydantic import BaseModel, Field, validator

from marvin._compat import PYDANTIC_V2, model_dump
from marvin.core.ChatCompletion.providers.openai import CONTEXT_SIZES
from marvin.openai import ChatCompletion
from marvin.prompts import library as prompt_library
from marvin.prompts.base import Prompt, render_prompts
from marvin.tools import Tool
from marvin.utilities.async_utils import run_sync
from marvin.utilities.history import History, HistoryFilter
from marvin.utilities.messages import Message, Role
from marvin.utilities.types import LoggerMixin, MarvinBaseModel

SYSTEM_PROMPT = """
    # Overview
    
    You are the intelligent, natural language interface to an application. The
    application has a structured `state` but no formal API; you are the only way
    to interact with it. You must interpret the user's inputs as attempts to
    interact with the application's state in the context of the application's
    purpose. For example, if the application is a to-do tracker, then "I need to
    go to the store" should be interpreted as an attempt to add a new to-do
    item. If it is a route planner, then "I need to go to the store" should be
    interpreted as an attempt to find a route to the store. 
    
    # Instructions
    
    Your primary job is to maintain the application's `state` and your own
    `plan`. Together, these two states fully parameterize the application,
    making it resilient, serializable, and observable. You do this autonomously;
    you do not need to inform the user of any changes you make. 
    
    # Actions
    
    Each time the user runs the application by sending a message, you must take
    the following steps:
    
    {% if app.plan_enabled %}

    - Call the `update_plan` function to update your plan. Use your plan
    to track notes, objectives, in-progress work, and to break problems down
    into solvable, possibly dependent parts. You plan consists of a few fields:
    
        - `notes`: a list of notes you have taken. Notes are free-form text and
        can be used to track anything you want to remember, such as
        long-standing user instructions, or observations about how to behave or
        operate the application. Your notes should always impact your behavior.
        These are exclusively related to your role as intermediary and you
        interact with the user and application. Do not track application data or
        state here.
        
        - `tasks`: a list of tasks you are working on. Tasks track goals,
        milestones, in-progress work, and break problems down into all the
        discrete steps needed to solve them. You should create a new task for
        any work that will require a function call other than updating state, or
        will require more than one state update to complete. You do not need to
        create tasks for simple state updates. Use optional parent tasks to
        indicate nested relationships; parent tasks are not completed until all
        their children are complete. Use optional upstream tasks to indicate
        dependencies; a task can not be completed until its upstream tasks are
        completed.
    
    {% endif %}
    
    - Call any functions necessary to achieve the application's purpose.

    {% if app.state_enabled %}

    - Call the `update_state` function to update the application's state. This
    is where you should store any information relevant to the application
    itself.
    
    {% endif %}
     
    You can call these functions at any time, in any order, as necessary.
    Finally, respond to the user with an informative message. Remember that the
    user is probably uninterested in the internal steps you took, so respond
    only in a manner appropriate to the application's purpose.

    # Application details
    
    ## Name
    
    {{ app.name }}
    
    ## Description
    
    {{ app.description or '' | render }}
    
    {% if app.state_enabled %}

    ## Application state
    
    {{ app.state.json() }}
    
    ### Application state schema
    
    {{ app.state.schema_json() }}
    
    {% endif %}
    
    {%- if app.plan_enabled %}
    
    ## AI (your) state
    
    {{ app.plan.json() }}
    
    ### AI state schema
    
    {{ app.plan.schema_json() }}
    
    {%- endif %}
    """


class TaskState(Enum):
    """The state of a task.

    Attributes:
        PENDING: The task is pending and has not yet started.
        IN_PROGRESS: The task is in progress.
        COMPLETED: The task is completed.
        FAILED: The task failed.
        SKIPPED: The task was skipped.
    """

    PENDING = "PENDING"
    IN_PROGRESS = "IN_PROGRESS"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"


class Task(BaseModel):
    class Config:
        validate_assignment = True

    id: int
    description: str
    upstream_task_ids: Optional[list[int]] = None
    parent_task_id: Optional[int] = None
    state: TaskState = TaskState.IN_PROGRESS


class AppPlan(BaseModel):
    """The AI's plan in service of the application.

    Attributes:
        tasks: A list of tasks the AI is working on.
        notes: A list of notes the AI has taken.
    """

    tasks: list[Task] = Field(default_factory=list)
    notes: list[str] = Field(default_factory=list)


class FreeformState(BaseModel):
    """A freeform state object that can be used to store any JSON-serializable data.

    Attributes:
        state: The state object.
    """

    state: dict[str, Any] = Field(default_factory=dict)


class AIApplication(LoggerMixin, MarvinBaseModel):
    """An AI application is a stateful, autonomous, natural language
        interface to an application.

    Attributes:
        name: The name of the application.
        description: A description of the application.
        state: The application's state - this can be any JSON-serializable object.
        plan: The AI's plan in service of the application - this can be any
            JSON-serializable object.
        tools: A list of tools that the AI can use to interact with
            application or outside world.
        history: A history of all messages sent and received by the AI.
        additional_prompts: A list of additional prompts that will be
            added to the prompt stack for rendering.

    Example:
        Create a simple todo app where AI manages its own state and plan.
        ```python
        from marvin import AIApplication

        todo_app = AIApplication(
            name="Todo App",
            description="A simple todo app.",
        )

        todo_app("I need to go to the store.")

        print(todo_app.state, todo_app.plan)
        ```
    """

    name: Optional[str] = None
    description: Optional[str] = None
    state: BaseModel = Field(default_factory=FreeformState)
    plan: AppPlan = Field(default_factory=AppPlan)
    tools: list[Union[Tool, Callable[..., Any]]] = Field(default_factory=list)
    history: History = Field(default_factory=History)
    additional_prompts: list[Prompt] = Field(
        default_factory=list,
        description=(
            "Additional prompts that will be added to the prompt stack for rendering."
        ),
    )
    stream_handler: Optional[Callable[[Message], None]] = None
    state_enabled: bool = True
    plan_enabled: bool = True

    @validator("description")
    def validate_description(cls, v):
        return inspect.cleandoc(v)

    @validator("additional_prompts")
    def validate_additional_prompts(cls, v):
        if v is None:
            v = []
        return v

    @validator("tools", pre=True, always=True)
    def validate_tools(cls, v):
        if v is None:
            v = []

        tools = []

        # convert AI Applications and functions to tools
        for tool in v:
            if isinstance(tool, (AIApplication, Tool)):
                tools.append(tool.as_function())
            elif callable(tool):
                tools.append(tool)
            else:
                raise ValueError(f"Tool {tool} is not a `Tool` or callable.")
        return tools

    @validator("name", always=True)
    def validate_name(cls, v):
        if v is None:
            v = cls.__name__
        return v

    def __call__(self, input_text: str = None, model: str = None):
        return run_sync(self.run(input_text=input_text, model=model))

    async def entrypoint(self, q: str) -> str:
        response = await self.run(input_text=q)
        return response.content

    async def run(self, input_text: str = None, model: str = None) -> Message:
        if model is None:
            model = "gpt-3.5-turbo"

        # set up prompts
        prompts = [
            # system prompts
            prompt_library.System(content=SYSTEM_PROMPT),
            # add current datetime
            prompt_library.Now(),
            # get the history of messages between user and assistant
            prompt_library.MessageHistory(
                history=self.history,
                skip=1,
                filter=HistoryFilter(role_in=[Role.USER, Role.ASSISTANT]),
            ),
            # get the user's latest input with higher priority the history
            prompt_library.User(content="{{ input_text }}"),
            *self.additional_prompts,
        ]

        message_list = render_prompts(
            prompts=prompts,
            render_kwargs=dict(app=self, input_text=input_text),
            max_tokens=CONTEXT_SIZES.get(model, 2048),
        )

        # get latest user input
        input_text = input_text or ""
        self.logger.debug_kv("User input", input_text, key_style="green")

        # set up tools
        tools = self.tools.copy()
        if self.state_enabled:
            tools.append(UpdateState(app=self).as_function())
        if self.plan_enabled:
            tools.append(UpdatePlan(app=self).as_function())

        conversation = await ChatCompletion(
            model=model,
            functions=tools,
            stream_handler=self.stream_handler,
        ).achain(messages=message_list)

        new_messages = [
            msg
            for msg in conversation.history
            if msg not in self.history.get_messages()
        ]

        for msg in new_messages:
            self.history.add_message(msg)

        self.logger.debug_kv("AI response", new_messages[-1].content, key_style="blue")
        return new_messages[-1]

    def as_tool(self, name: str = None) -> Tool:
        return AIApplicationTool(app=self, name=name)

    def as_function(self, name: str = None) -> Callable:
        return self.as_tool(name=name).as_function()


class AIApplicationTool(Tool):
    app: "AIApplication"

    def __init__(self, **kwargs):
        if "name" not in kwargs:
            kwargs["name"] = type(self.app).__name__
        super().__init__(**kwargs)

    def run(self, input_text: str) -> str:
        return run_sync(self.app.run(input_text))


class JSONPatchModel(
    BaseModel,
    **(
        {
            "allow_population_by_field_name": True,
        }
        if not PYDANTIC_V2
        else {
            "populate_by_name": True,
        }
    ),
):
    """A JSON Patch document.

    Attributes:
        op: The operation to perform.
        path: The path to the value to update.
        value: The value to update the path to.
        from_: The path to the value to copy from.
    """

    op: str
    path: str
    value: Union[str, float, int, bool, list, dict] = None
    from_: str = Field(None, alias="from")


class UpdateState(Tool):
    """A `Tool` that updates the apps state using JSON Patch documents.

    Example:
        Manually update the state of an AI Application.
        ```python
        from marvin.components.ai_application import (
            AIApplication,
            FreeformState,
            JSONPatchModel,
            UpdateState,
        )

        destination_tracker = AIApplication(
            name="Destination Tracker",
            description="keeps track of where i've been",
            state=FreeformState(state={"San Francisco": "not visited"}),
        )

        UpdateState(app=destination_tracker).run([
            {
                "op": "replace",
                "path": "/state/San Francisco",
                "value": "visited"
            }
        ])

        assert destination_tracker.state.dict() == {
            "state": {"San Francisco": "visited"}
        }
        ```
    """

    app: "AIApplication" = Field(..., repr=False, exclude=True)
    description: str = """
        Update the application state by providing a list of JSON patch
        documents. The state must always comply with the state's
        JSON schema.
        """

    def __init__(self, app: AIApplication, **kwargs):
        super().__init__(**kwargs, app=app)

    def run(self, patches: list[JSONPatchModel]):
        patch = JsonPatch(patches)
        updated_state = patch.apply(model_dump(self.app.state))
        self.app.state = type(self.app.state)(**updated_state)
        return "Application state updated successfully!"


class UpdatePlan(Tool):
    """
    A `Tool` that updates the apps plan using JSON Patch documents.


    Example:
        Manually update task status in an AI Application's plan.
        ```python
        from marvin.components.ai_application import (
            AIApplication,
            AppPlan,
            UpdatePlan,
        )

        todo_app = AIApplication(name="Todo App", description="A simple todo app")

        todo_app("i need to buy milk")

        # manually update the plan (usually done by the AI)
        UpdatePlan(app=todo_app).run([
            {
                "op": "replace",
                "path": "/tasks/0/state",
                "value": "COMPLETED"
            }
        ])

        print(todo_app.plan)
        ```
    """

    app: "AIApplication" = Field(..., repr=False, exclude=True)
    description: str = """
        Update the application plan by providing a list of JSON patch
        documents. The state must always comply with the plan's JSON schema.
        """

    def __init__(self, app: AIApplication, **kwargs):
        super().__init__(**kwargs, app=app)

    def run(self, patches: list[JSONPatchModel]):
        patch = JsonPatch(patches)

        updated_plan = patch.apply(model_dump(self.app.plan))
        self.app.plan = type(self.app.plan)(**updated_plan)
        return "Application plan updated successfully!"



from types import FunctionType, GenericAlias
from typing import (
    Annotated,
    Any,
    Callable,
    Optional,
    TypeVar,
    Union,
    cast,
    get_origin,
)

from pydantic import BaseModel, create_model
from pydantic.version import VERSION as PYDANTIC_VERSION

_ModelT = TypeVar("_ModelT", bound="BaseModel")

PYDANTIC_V2 = PYDANTIC_VERSION.startswith("2.")

if PYDANTIC_V2:
    from pydantic.v1 import (
        BaseSettings,
        Field,
        SecretStr,
        validate_arguments,
    )

    SettingsConfigDict = BaseSettings.Config

    from pydantic import field_validator  # noqa # type: ignore

else:
    from pydantic import (  # noqa # type: ignore
        BaseSettings,
        Field,
        SecretStr,
        validate_arguments,
        validator as field_validator,
    )

    SettingsConfigDict = BaseSettings.Config


def model_dump(model: _ModelT, **kwargs: Any) -> dict[str, Any]:
    if PYDANTIC_V2 and hasattr(model, "model_dump"):
        return model.model_dump(**kwargs)  # type: ignore
    return model.dict(**kwargs)  # type: ignore


def model_dump_json(model: type[BaseModel], **kwargs: Any) -> dict[str, Any]:
    if PYDANTIC_V2 and hasattr(model, "model_dump_json"):
        return model.model_dump_json(**kwargs)  # type: ignore
    return model.json(**kwargs)  # type: ignore


def model_json_schema(
    model: type[BaseModel],
    name: Optional[str] = None,
    description: Optional[str] = None,
) -> dict[str, Any]:
    # Get the schema from the model.
    schema = {"parameters": {**model_schema(model)}}

    # Mutate the schema to match the OpenAPI spec.
    schema["parameters"]["title"] = name or schema["parameters"].pop("title", None)
    schema["parameters"]["description"] = description or schema["parameters"].pop(
        "description", ""
    )  # noqa

    # Move the properties to the root of the schema.
    schema["name"] = schema["parameters"].pop("title")
    schema["description"] = schema["parameters"].pop("description")
    return schema


def model_schema(model: type[BaseModel], **kwargs: Any) -> dict[str, Any]:
    if PYDANTIC_V2 and hasattr(model, "model_json_schema"):
        return model.model_json_schema(**kwargs)  # type: ignore
    return model.schema(**kwargs)  # type: ignore


def model_copy(model: _ModelT, **kwargs: Any) -> _ModelT:
    if PYDANTIC_V2 and hasattr(model, "model_copy"):
        return model.model_copy(**kwargs)  # type: ignore
    return model.copy(**kwargs)  # type: ignore


def cast_callable_to_model(
    function: Callable[..., Any],
    name: Optional[str] = None,
    description: Optional[str] = None,
) -> type[BaseModel]:
    response = validate_arguments(function).model  # type: ignore
    for field in ["args", "kwargs", "v__duplicate_kwargs"]:
        fields = cast(dict[str, Any], response.__fields__)  # type: ignore
        fields.pop(field, None)
    response.__title__ = name or function.__name__
    response.__name__ = name or function.__name__
    response.__doc__ = description or function.__doc__
    return response  # type: ignore


def cast_type_or_alias_to_model(
    type_: Union[type, GenericAlias],
    name: Optional[str] = None,
    description: Optional[str] = None,
    field_name: Optional[str] = None,
) -> type[BaseModel]:
    fields: dict[str, Any] = {}
    fields[field_name or "output"] = (type_, ...)
    response = create_model(
        name or "Output",
        __base__=BaseModel,
        **fields,
    )
    response.__doc__ = description or response.__doc__
    return response


def cast_to_model(
    function_or_type: Union[type, type[BaseModel], GenericAlias, Callable[..., Any]],
    name: Optional[str] = None,
    description: Optional[str] = None,
    field_name: Optional[str] = None,
) -> type[BaseModel]:
    """
    Casts a type or callable to a Pydantic model.
    """
    response = BaseModel
    if get_origin(function_or_type) is Annotated:
        metadata: Any = next(iter(function_or_type.__metadata__), None)  # type: ignore
        annotated_field_name: Optional[str] = field_name

        if hasattr(metadata, "extra") and isinstance(metadata.extra, dict):
            annotated_field_name: Optional[str] = metadata.extra.get("name", "")  # type: ignore # noqa
        elif hasattr(metadata, "json_schema_extra") and isinstance(
            metadata.json_schema_extra, dict
        ):  # noqa
            annotated_field_name: Optional[str] = metadata.json_schema_extra.get("name", "")  # type: ignore # noqa
        elif isinstance(metadata, dict):
            annotated_field_name: Optional[str] = metadata.get("name", "")  # type: ignore # noqa
        elif isinstance(metadata, str):
            annotated_field_name: Optional[str] = metadata
        else:
            pass
        annotated_field_description: Optional[str] = description or ""
        if hasattr(metadata, "description") and isinstance(metadata.description, str):
            annotated_field_description: Optional[str] = metadata.description  # type: ignore # noqa
        elif isinstance(metadata, dict):
            annotated_field_description: Optional[str] = metadata.get("description", "")  # type: ignore # noqa
        else:
            pass

        response = cast_to_model(
            function_or_type.__origin__,  # type: ignore
            name=name,
            description=annotated_field_description,
            field_name=annotated_field_name,  # type: ignore
        )
        response.__doc__ = annotated_field_description or ""
    if isinstance(function_or_type, GenericAlias):
        response = cast_type_or_alias_to_model(
            function_or_type, name, description, field_name
        )
    elif isinstance(function_or_type, type):
        if issubclass(function_or_type, BaseModel):
            response = create_model(
                name or function_or_type.__name__,
                __base__=function_or_type,
            )
            response.__doc__ = description or function_or_type.__doc__

        else:
            response = cast_type_or_alias_to_model(
                function_or_type, name, description, field_name
            )
    elif isinstance(function_or_type, Callable):
        if isinstance(function_or_type, FunctionType):
            response = cast_callable_to_model(function_or_type, name, description)
    else:
        response = cast_type_or_alias_to_model(
            function_or_type, name, description, field_name
        )
    return response


def cast_to_json(
    function_or_type: Union[type, type[BaseModel], GenericAlias, Callable[..., Any]],
    name: Optional[str] = None,
    description: Optional[str] = None,
    field_name: Optional[str] = None,
) -> dict[str, Any]:
    return model_json_schema(
        cast_to_model(function_or_type, name, description, field_name)
    )



import inspect
import re
from datetime import datetime
from zoneinfo import ZoneInfo

import tiktoken
from jinja2 import (
    ChoiceLoader,
    Environment,
    StrictUndefined,
    pass_context,
    select_autoescape,
)
from markupsafe import Markup

import marvin.utilities.async_utils

NEWLINES_REGEX = re.compile(r"(\s*\n\s*)")
MD_LINK_REGEX = r"\[(?P<text>[^\]]+)]\((?P<url>[^\)]+)\)"

jinja_env = Environment(
    loader=ChoiceLoader(
        [
            # PackageLoader("marvin", "prompts")
        ]
    ),
    autoescape=select_autoescape(default_for_string=False),
    trim_blocks=True,
    lstrip_blocks=True,
    auto_reload=True,
    undefined=StrictUndefined,
)

jinja_env.globals.update(
    zip=zip,
    arun=marvin.utilities.async_utils.run_sync,
    now=lambda: datetime.now(ZoneInfo("UTC")),
)


@pass_context
def render_filter(context, value):
    """
    Allows nested rendering of variables that may contain variables themselves
    e.g. {{ description | render }}
    """
    _template = context.eval_ctx.environment.from_string(value)
    result = _template.render(**context)
    if context.eval_ctx.autoescape:
        result = Markup(result)
    return result


jinja_env.filters["render"] = render_filter


def tokenize(text: str) -> list[int]:
    tokenizer = tiktoken.encoding_for_model("gpt-3.5-turbo")
    return tokenizer.encode(text)


def detokenize(tokens: list[int]) -> str:
    tokenizer = tiktoken.encoding_for_model("gpt-3.5-turbo")
    return tokenizer.decode(tokens)


def count_tokens(text: str) -> int:
    return len(tokenize(text))


def slice_tokens(text: str, n_tokens: int) -> str:
    tokens = tokenize(text)
    return detokenize(tokens[:n_tokens])


def split_tokens(text: str, n_tokens: int) -> list[str]:
    tokens = tokenize(text)
    return [
        detokenize(tokens[i : i + n_tokens]) for i in range(0, len(tokens), n_tokens)
    ]


def condense_newlines(text: str) -> str:
    def replace_whitespace(match):
        newlines_count = match.group().count("\n")
        if newlines_count <= 1:
            return " "
        else:
            return "\n" * newlines_count

    text = inspect.cleandoc(text)
    text = NEWLINES_REGEX.sub(replace_whitespace, text)
    return text.strip()


def html_to_content(html: str) -> str:
    from bs4 import BeautifulSoup

    soup = BeautifulSoup(html, "html.parser")

    # Remove script and style elements
    for script in soup(["script", "style"]):
        script.extract()

    # Get text
    text = soup.get_text()

    # Condense newlines
    return condense_newlines(text)


def convert_md_links_to_slack(text) -> str:
    # converting Markdown links to Slack-style links
    def to_slack_link(match):
        return f'<{match.group("url")}|{match.group("text")}>'

    # Replace Markdown links with Slack-style links
    slack_text = re.sub(MD_LINK_REGEX, to_slack_link, text)

    return slack_text


def split_text_by_tokens(text: str, split_tokens: list[str]) -> list[tuple[str, str]]:
    cleaned_text = inspect.cleandoc(text)

    # Find all positions of tokens in the text
    positions = [
        (match.start(), match.end(), match.group().rstrip(":").strip())
        for token in split_tokens
        for match in re.finditer(re.escape(token) + r"(?::\s*)?", cleaned_text)
    ]

    # Sort positions by their start index
    positions.sort(key=lambda x: x[0])

    paired: list[tuple[str, str]] = []
    prev_end = 0
    prev_token = split_tokens[0]
    for start, end, token in positions:
        paired.append((prev_token, cleaned_text[prev_end:start].strip()))
        prev_end = end
        prev_token = token

    paired.append((prev_token, cleaned_text[prev_end:].strip()))

    # Remove pairs where the text is empty
    paired = [(token.replace(":", ""), text) for token, text in paired if text]

    return paired



import inspect
from datetime import date, datetime, timedelta

import httpx
import marvin
from marvin import ai_fn
from marvin.utilities.strings import jinja_env
from prefect import flow, task
from prefect.artifacts import create_markdown_artifact
from prefect.blocks.system import Secret

DAILY_DIGEST_TEMPLATE = jinja_env.from_string(inspect.cleandoc("""
        # GitHub Digest: {{ today }}
        
        Hi {{ username }}! Here's what you've been up to on GitHub today:
        
        ## you created {{ created_issues|length }} issue(s) and opened {{ created_pull_requests|length }} PR(s):
        ### issues
        {% for issue in created_issues %}
        - [{{ issue.title }}]({{ issue.html_url }})
        {% endfor %}
        
        ### PRs
        {% for pr in created_pull_requests %}
        - [{{ pr.title }}]({{ pr.html_url }})
        {% endfor %}
        
        ## you closed {{ closed_issues|length }} issue(s) and merged {{ closed_pull_requests|length }} PR(s):
        ### issues
        {% for issue in closed_issues %}
        - [{{ issue.title }}]({{ issue.html_url }})
        {% endfor %}

        ### PRs
        {% for pr in closed_pull_requests %}
        - [{{ pr.title }}]({{ pr.html_url }})
        {% endfor %}

        ## you committed to {{ committed_repos|length }} repo(s):
        {% for repo in committed_repos %}
        - {{ repo.commits|length }} commits to [{{ repo.name }}](github.com/{{ repo.name }})
        {% endfor %}
        """))  # noqa: E501


@ai_fn
async def summarize_digest(markdown_digest: str) -> str:
    """Produce a short story based on the GitHub digest.

    The subject of the story is the username greeted in the beginning of the digest.
    """


@task
async def get_digest_data(username, gh_token_secret_name, since, max=100):
    events_url = f"https://api.github.com/users/{username}/events/public?per_page={max}"

    token = await Secret.load(gh_token_secret_name)

    created_issues = []
    created_pull_requests = []
    closed_issues = []
    closed_pull_requests = []
    committed_repos = {}

    async with httpx.AsyncClient(
        headers={
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"token {token.get()}",
        }
    ) as client:
        for event in (await client.get(events_url)).json():
            created_at = datetime.fromisoformat(event["created_at"].rstrip("Z"))
            if created_at < since:
                continue

            if event["type"] == "IssuesEvent":
                issue = event["payload"]["issue"]
                if event["payload"]["action"] == "opened":
                    created_issues.append(issue)
                elif event["payload"]["action"] == "closed":
                    closed_issues.append(issue)

            elif event["type"] == "PullRequestEvent":
                pr = event["payload"]["pull_request"]
                if event["payload"]["action"] == "opened":
                    created_pull_requests.append(pr)
                elif event["payload"]["action"] == "closed":
                    closed_pull_requests.append(pr)

            elif event["type"] == "PushEvent":
                repo_name = event["repo"]["name"]
                commits = event["payload"]["commits"]
                if repo_name not in committed_repos:
                    committed_repos[repo_name] = {
                        "name": repo_name,
                        "commits": [],
                    }
                for commit_data in commits:
                    commit = (await client.get(commit_data["url"])).json()
                    if (
                        "Merge remote-tracking branch"
                        not in commit["commit"]["message"]
                    ):
                        committed_repos[repo_name]["commits"].append(commit)

    digest_data = {
        "today": date.today().strftime("%Y-%m-%d"),
        "created_issues": created_issues,
        "created_pull_requests": created_pull_requests,
        "closed_issues": closed_issues,
        "closed_pull_requests": closed_pull_requests,
        "committed_repos": list(committed_repos.values()),
    }

    return digest_data


@flow
async def daily_github_digest(username: str, gh_token_secret_name: str):
    """
    A flow that creates a daily digest of GitHub activity.

    Args:
        username: The GitHub username to create a digest for.
        gh_token_secret_name: The name of the secret containing a GitHub token.
    """
    since = datetime.utcnow() - timedelta(days=1)

    data = await get_digest_data(username, gh_token_secret_name, since)

    markdown_digest = DAILY_DIGEST_TEMPLATE.render(username=username, **data)

    tldr = await summarize_digest(markdown_digest)

    await create_markdown_artifact(
        key="github-digest",
        markdown=markdown_digest,
        description=tldr,
    )


if __name__ == "__main__":
    import asyncio

    marvin.settings.llm_model = "gpt-4"

    asyncio.run(
        daily_github_digest(username="zzstoatzz", gh_token_secret_name="github-token")
    )



import logging
from functools import lru_cache, partial

from rich.logging import RichHandler
from rich.markup import escape

import marvin


@lru_cache()
def get_logger(name: str = None) -> logging.Logger:
    parent_logger = logging.getLogger("marvin")

    if name:
        # Append the name if given but allow explicit full names e.g. "marvin.test"
        # should not become "marvin.marvin.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    add_logging_methods(logger)
    return logger


def setup_logging(level: str = None):
    logger = get_logger()

    if level is not None:
        logger.setLevel(level)
    else:
        logger.setLevel(marvin.settings.log_level)

    if not any(isinstance(h, RichHandler) for h in logger.handlers):
        handler = RichHandler(
            rich_tracebacks=True,
            markup=False,
            # console=Console(width=marvin.settings.log_console_width),
        )
        formatter = logging.Formatter("%(name)s: %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)


def add_logging_methods(logger):
    def log_style(level: int, message: str, style: str = None):
        if not style:
            style = "default on default"
        message = f"[{style}]{escape(str(message))}[/]"
        logger.log(level, message, extra={"markup": True})

    def log_kv(
        level: int,
        key: str,
        value: str,
        key_style: str = "default on default",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ):
        logger.log(
            level,
            f"[{key_style}]{escape(str(key))}{delimiter}[/][{value_style}]{escape(str(value))}[/]",
            extra={"markup": True},
        )

    logger.debug_style = partial(log_style, logging.DEBUG)
    logger.info_style = partial(log_style, logging.INFO)
    logger.warning_style = partial(log_style, logging.WARNING)
    logger.error_style = partial(log_style, logging.ERROR)
    logger.critical_style = partial(log_style, logging.CRITICAL)

    logger.debug_kv = partial(log_kv, logging.DEBUG)
    logger.info_kv = partial(log_kv, logging.INFO)
    logger.warning_kv = partial(log_kv, logging.WARNING)
    logger.error_kv = partial(log_kv, logging.ERROR)
    logger.critical_kv = partial(log_kv, logging.CRITICAL)


setup_logging(level=marvin.settings.log_level)



import secrets
import string


def create_secure_key(length=50):
    alphabet = string.ascii_letters + string.digits + "_"
    secure_key = "".join(secrets.choice(alphabet) for _ in range(length))
    return secure_key



from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar

from marvin._compat import model_copy, model_dump
from marvin.utilities.messages import Message
from pydantic import BaseModel, Field
from typing_extensions import Self

from .handlers import Request, Response, Turn

T = TypeVar(
    "T",
    bound=BaseModel,
)


class Conversation(BaseModel, Generic[T], extra="allow", arbitrary_types_allowed=True):
    turns: list[Turn[T]]
    model: Any

    def __getitem__(self, key: int) -> Turn[T]:
        return self.turns[key]

    @property
    def last_turn(self) -> Turn[T]:
        return self.turns[-1]

    @property
    def last_request(self) -> Optional[Request[T]]:
        return self.turns[-1][0] if self.turns else None

    @property
    def last_response(self) -> Optional[Response[T]]:
        return self.turns[-1][1] if self.turns else None

    @property
    def history(self) -> list[Message]:
        response: list[Message] = []
        if not self.turns:
            return response
        if self.last_request:
            response = self.last_request.messages or []
            if self.last_response:
                response.append(self.last_response.choices[0].message)
        return response

    def send(self, messages: list[Message], **kwargs: Any) -> Turn[T]:
        params = kwargs
        if self.last_request:
            params = model_dump(self.last_request, exclude={"messages"}) | kwargs

        turn = self.model.create(
            **params,
            messages=[
                *self.history,
                *messages,
            ],
        )
        self.turns.append(turn)
        return turn

    async def asend(self, messages: list[Message], **kwargs: Any) -> Turn[T]:
        params = kwargs
        if self.last_request:
            params = model_dump(self.last_request, exclude={"messages"}) | kwargs

        turn = await self.model.acreate(
            **params,
            messages=[
                *self.history,
                *messages,
            ],
        )
        self.turns.append(turn)
        return turn


class AbstractChatCompletion(
    BaseModel, Generic[T], ABC, extra="allow", arbitrary_types_allowed=True
):
    """
    A ChatCompletion object is responsible for exposing a create and acreate method,
    and for merging default parameters with the parameters passed to these methods.
    """

    defaults: dict[str, Any] = Field(default_factory=dict, exclude=True)

    def __call__(self: Self, **kwargs: Any) -> Self:
        """
        Create a new ChatCompletion object with new defaults computed from
        merging the passed parameters with the default parameters.
        """
        copy = model_copy(self)
        copy.defaults = self.defaults | kwargs
        return copy

    @abstractmethod
    def _serialize_request(self, request: Optional[Request[T]]) -> dict[str, Any]:
        """
        Serialize the request.
        This should be implemented by derived classes based on their specific needs.
        """
        pass

    @abstractmethod
    def _create_request(self, **kwargs: Any) -> Request[T]:
        """
        Prepare and return a request object.
        This should be implemented by derived classes.
        """
        pass

    @abstractmethod
    def _parse_response(self, response: Any) -> Any:
        """
        Parse the response based on specific needs.
        """
        pass

    def merge_with_defaults(self, **kwargs: Any) -> dict[str, Any]:
        """
        Merge the passed parameters with the default parameters.
        """
        return self.defaults | kwargs

    @abstractmethod
    def _send_request(self, **serialized_request: Any) -> Any:
        """
        Send the serialized request to the appropriate endpoint/service.
        Derived classes should implement this.
        """
        pass

    @abstractmethod
    async def _send_request_async(
        self, **serialized_request: Any
    ) -> Response[T]:  # noqa
        """
        Send the serialized request to the appropriate endpoint/service asynchronously.
        Derived classes should implement this.
        """
        pass

    def create(self, response_model: Optional[type] = None, **kwargs: Any) -> Turn[T]:
        """
        Create a completion synchronously.
        Derived classes can override this if they need to change the core logic.
        """
        request = self._create_request(**kwargs, response_model=response_model)
        serialized_request = self._serialize_request(request=request)
        response_data = self._send_request(**serialized_request)
        response = self._parse_response(response_data)

        return Turn(
            request=Request(
                **serialized_request
                | self.defaults
                | model_dump(request, exclude_none=True)
                | ({"response_model": response_model} if response_model else {})
            ),
            response=response,
        )

    async def acreate(
        self, response_model: Optional[type[T]] = None, **kwargs: Any
    ) -> Turn[T]:
        """
        Create a completion asynchronously.
        Similar to the synchronous version but for async implementations.
        """
        request = self._create_request(**kwargs, response_model=response_model)
        serialized_request = self._serialize_request(request=request)
        response_data = await self._send_request_async(**serialized_request)
        response = self._parse_response(response_data)
        return Turn(
            request=Request(
                **serialized_request
                | self.defaults
                | model_dump(request, exclude_none=True)
                | ({"response_model": response_model} if response_model else {})
            ),
            response=response,
        )

    def chain(self, **kwargs: Any) -> Conversation[T]:
        """
        Create a new Conversation object.
        """
        with self as conversation:
            conversation.send(**kwargs)
            while conversation.last_turn.has_function_call():
                message = conversation.last_turn.call_function()
                conversation.send(
                    message if isinstance(message, list) else [message],
                )

            return conversation

    async def achain(self, **kwargs: Any) -> Conversation[T]:
        """
        Create a new Conversation object asynchronously.
        """
        with self as conversation:
            await conversation.asend(**kwargs)
            while conversation.last_turn.has_function_call():
                message = conversation.last_turn.call_function()
                await conversation.asend(
                    message if isinstance(message, list) else [message],
                )

            return conversation

    def __enter__(self: Self) -> Conversation[T]:
        """
        Enter a context manager.
        """
        return Conversation(turns=[], model=self)

    def __exit__(self: Self, *args: Any) -> None:
        """
        Exit a context manager.
        """
        pass



import asyncio
import json
from typing import Optional

import httpx
from typing_extensions import Literal

import marvin
from marvin.tools import Tool
from marvin.utilities.embeddings import create_openai_embeddings

QueryResultType = Literal["documents", "distances", "metadatas"]


async def list_collections() -> list[dict]:
    async with httpx.AsyncClient() as client:
        chroma_api_url = f"http://{marvin.settings.chroma_server_host}:{marvin.settings.chroma_server_http_port}"
        response = await client.get(
            f"{chroma_api_url}/api/v1/collections",
        )

    response.raise_for_status()
    return response.json()


async def query_chroma(
    query: str,
    collection: str = "marvin",
    n_results: int = 5,
    where: Optional[dict] = None,
    where_document: Optional[dict] = None,
    include: Optional[list[QueryResultType]] = None,
    max_characters: int = 2000,
) -> str:
    query_embedding = (await create_openai_embeddings([query]))[0]

    collection_ids = [
        c["id"] for c in await list_collections() if c["name"] == collection
    ]

    if len(collection_ids) == 0:
        return f"Collection {collection} not found."

    collection_id = collection_ids[0]

    async with httpx.AsyncClient() as client:
        chroma_api_url = f"http://{marvin.settings.chroma_server_host}:{marvin.settings.chroma_server_http_port}"

        response = await client.post(
            f"{chroma_api_url}/api/v1/collections/{collection_id}/query",
            data=json.dumps(
                {
                    "query_embeddings": [query_embedding],
                    "n_results": n_results,
                    "where": where or {},
                    "where_document": where_document or {},
                    "include": include or ["documents"],
                }
            ),
            headers={"Content-Type": "application/json"},
        )

    response.raise_for_status()

    return "\n".join(
        [
            f"{i+1}. {', '.join(excerpt)}"
            for i, excerpt in enumerate(response.json()["documents"])
        ]
    )[:max_characters]


class QueryChroma(Tool):
    """Tool for querying a Chroma index."""

    description: str = """
        Retrieve document excerpts from a knowledge-base given a query.
    """

    async def run(
        self,
        query: str,
        collection: str = "marvin",
        n_results: int = 5,
        where: Optional[dict] = None,
        where_document: Optional[dict] = None,
        include: Optional[list[QueryResultType]] = None,
        max_characters: int = 2000,
    ) -> str:
        return await query_chroma(
            query, collection, n_results, where, where_document, include, max_characters
        )


class MultiQueryChroma(Tool):
    """Tool for querying a Chroma index."""

    description: str = """
        Retrieve document excerpts from a knowledge-base given a query.
    """

    async def run(
        self,
        queries: list[str],
        collection: str = "marvin",
        n_results: int = 5,
        where: Optional[dict] = None,
        where_document: Optional[dict] = None,
        include: Optional[list[QueryResultType]] = None,
        max_characters: int = 2000,
        max_queries: int = 5,
    ) -> str:
        if len(queries) > max_queries:
            # make sure excerpts are not too short
            queries = queries[:max_queries]

        coros = [
            query_chroma(
                query,
                collection,
                n_results,
                where,
                where_document,
                include,
                max_characters // len(queries),
            )
            for query in queries
        ]
        return "\n\n".join(await asyncio.gather(*coros, return_exceptions=True))



import asyncio
import inspect
from functools import partial
from typing import Any, Callable, Optional, TypeVar

from pydantic import BaseModel
from typing_extensions import ParamSpec, Self

from marvin.core.ChatCompletion import ChatCompletion
from marvin.core.ChatCompletion.abstract import AbstractChatCompletion
from marvin.prompts import Prompt, prompt_fn
from marvin.utilities.async_utils import run_sync

T = TypeVar("T", bound=BaseModel)

A = TypeVar("A", bound=Any)

P = ParamSpec("P")


def ai_model_prompt(
    cls: type[BaseModel],
    ctx: Optional[dict[str, Any]] = None,
    **kwargs: Any,
) -> Callable[[str], Prompt[P]]:
    description = cls.__doc__ or ""
    if ctx and ctx.get("instructions") and isinstance(ctx.get("instructions"), str):
        instructions = str(ctx.get("instructions"))
        description += "\n" + instructions if (instructions != description) else ""

    @prompt_fn(
        ctx={"ctx": ctx or {}, "inspect": inspect},
        response_model=cls,
        response_model_name="FormatResponse",
        response_model_description=description,
        serialize_on_call=False,
    )
    def prompt_wrapper(text: str) -> None:  # type: ignore # noqa
        """
        The user will provide text that you need to parse into a
        structured form {{'(note you must also: ' + ctx.get('instructions') + ')' if ctx.get('instructions')}}.
        To validate your response, you must call the
        `{{response_model.__name__}}` function.
        Use the provided text and context to extract, deduce, or infer
        any parameters needed by `{{response_model.__name__}}`, including any missing
        data.

        You have been provided the following context to perform your task:
            - The current time is {{now()}}.
        {% set context = ctx.get('context_fn')(text).items() if ctx.get('context_fn') %}
        {% if context %}
        {%for (arg, value) in context%}
            - {{ arg }}: {{ value }}\n
        {% endfor %}
        {% endif %}

        User: The text to parse: {{text}}


        """  # noqa

    return prompt_wrapper  # type: ignore


class AIModel(BaseModel):
    def __init__(
        self,
        text: Optional[str] = None,
        /,
        instructions_: Optional[str] = None,
        **kwargs: Any,
    ):
        if text:
            kwargs.update(self.__class__.call(text, instructions=instructions_))

        super().__init__(**kwargs)

    @classmethod
    def get_prompt(
        cls,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
    ) -> Callable[[str], Prompt[P]]:
        ctx = ctx or getattr(cls, "__metadata__", {}).get("ctx", {})
        instructions = (  # type: ignore
            "\n".join(
                list(
                    filter(
                        bool,
                        [
                            instructions,
                            getattr(cls, "__metadata__", {}).get("instructions"),
                        ],
                    )
                )  # type: ignore
            )
            or None
        )

        response_model_name = response_model_name or getattr(
            cls, "__metadata__", {}
        ).get("response_model_name")
        response_model_description = response_model_description or getattr(
            cls, "__metadata__", {}
        ).get("response_model_description")
        response_model_field_name = response_model_field_name or getattr(
            cls, "__metadata__", {}
        ).get("response_model_field_name")

        return ai_model_prompt(
            cls,
            ctx=((ctx or {}) | {"instructions": instructions}),
            response_model_name=response_model_name,
            response_model_description=response_model_description,
            response_model_field_name=response_model_field_name,
        )

    @classmethod
    def as_prompt(
        cls,
        text: str,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> dict[str, Any]:
        metadata = getattr(cls, "__metadata__", {})

        # Set default values using a loop to reduce repetition
        default_keys = [
            "ctx",
            "instructions",
            "response_model_name",
            "response_model_description",
            "response_model_field_name",
            "model",
            "model_kwargs",
        ]
        local_vars = locals()
        for key in default_keys:
            if local_vars.get(key, None) is None:
                local_vars[key] = metadata.get(key, {})

        return cls.get_prompt(
            ctx=ctx,
            instructions=instructions,
            response_model_name=response_model_name,
            response_model_description=response_model_description,
            response_model_field_name=response_model_field_name,
        )(text).serialize(model=ChatCompletion(model=model, **model_kwargs))

    @classmethod
    def as_dict(
        cls,
        text: str,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> dict[str, Any]:
        metadata = getattr(cls, "__metadata__", {})

        # Set default values using a loop to reduce repetition
        default_keys = [
            "ctx",
            "instructions",
            "response_model_name",
            "response_model_description",
            "response_model_field_name",
            "model",
            "model_kwargs",
        ]
        local_vars = locals()
        for key in default_keys:
            if local_vars.get(key, None) is None:
                local_vars[key] = metadata.get(key, {})
        return cls.get_prompt(
            ctx=ctx,
            instructions=instructions,
            response_model_name=response_model_name,
            response_model_description=response_model_description,
            response_model_field_name=response_model_field_name,
        )(text).to_dict()

    @classmethod
    def as_chat_completion(
        cls,
        text: str,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> AbstractChatCompletion[T]:  # type: ignore
        metadata = getattr(cls, "__metadata__", {})

        # Set default values using a loop to reduce repetition
        default_keys = [
            "ctx",
            "instructions",
            "response_model_name",
            "response_model_description",
            "response_model_field_name",
            "model",
            "model_kwargs",
        ]
        local_vars = locals()
        for key in default_keys:
            if local_vars.get(key, None) is None:
                local_vars[key] = metadata.get(key, {})

        return ChatCompletion(model=model, **model_kwargs)(
            **cls.as_dict(
                text,
                ctx=ctx,
                instructions=instructions,
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
            )
        )  # type: ignore

    @classmethod
    def call(
        cls: type[Self],
        text: str,
        *,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> Self:
        metadata = getattr(cls, "__metadata__", {})

        # Set default values using a loop to reduce repetition
        default_keys = [
            "ctx",
            "instructions",
            "response_model_name",
            "response_model_description",
            "response_model_field_name",
            "model",
            "model_kwargs",
        ]
        local_vars = locals()
        for key in default_keys:
            if local_vars.get(key, None) is None:
                local_vars[key] = metadata.get(key, {})

        _model: Self = (  # type: ignore
            cls.as_chat_completion(
                text,
                ctx=ctx,
                instructions=instructions,
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
                model=model,
                **model_kwargs,
            )
            .create()
            .to_model()
        )
        return _model  # type: ignore

    @classmethod
    async def acall(
        cls: type[Self],
        text: str,
        *,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> Self:
        metadata = getattr(cls, "__metadata__", {})

        # Set default values using a loop to reduce repetition
        default_keys = [
            "ctx",
            "instructions",
            "response_model_name",
            "response_model_description",
            "response_model_field_name",
            "model",
            "model_kwargs",
        ]
        local_vars = locals()
        for key in default_keys:
            if local_vars.get(key, None) is None:
                local_vars[key] = metadata.get(key, {})

        _model: Self = (  # type: ignore
            await cls.as_chat_completion(
                text,
                ctx=ctx,
                instructions=instructions,
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
                model=model,
                **model_kwargs,
            ).acreate()  # type: ignore
        ).to_model()
        return _model  # type: ignore

    @classmethod
    def map(cls, *map_args: list[str], **map_kwargs: list[Any]):
        """
        Map the AI function over a sequence of arguments. Runs concurrently.

        Arguments should be provided as if calling the function normally, but
        each argument must be a list. The function is called once for each item
        in the list, and the results are returned in a list.

        This method should be called synchronously.

        For example, fn.map([1, 2]) is equivalent to [fn(1), fn(2)].

        fn.map([1, 2], x=['a', 'b']) is equivalent to [fn(1, x='a'), fn(2, x='b')].
        """
        return run_sync(cls.amap(*map_args, **map_kwargs))

    @classmethod
    async def amap(cls, *map_args: list[str], **map_kwargs: list[Any]) -> list[Any]:
        tasks: list[Any] = []
        if map_args:
            max_length = max(len(arg) for arg in map_args)
        else:
            max_length = max(len(v) for v in map_kwargs.values())

        for i in range(max_length):
            call_args: list[str] = [
                arg[i] if i < len(arg) else None for arg in map_args
            ]  # type: ignore

            tasks.append(cls.acall(*call_args, **map_kwargs))

        return await asyncio.gather(*tasks)

    @classmethod
    def as_decorator(
        cls: type[Self],
        base_model: Optional[type[BaseModel]] = None,
        ctx: Optional[dict[str, Any]] = None,
        instructions: Optional[str] = None,
        response_model_name: Optional[str] = None,
        response_model_description: Optional[str] = None,
        response_model_field_name: Optional[str] = None,
        model: Optional[str] = None,
        **model_kwargs: Any,
    ) -> type[BaseModel]:
        if not base_model:
            return partial(
                cls.as_decorator,
                ctx=ctx,
                instructions=instructions,
                response_model_name=response_model_name,
                response_model_description=response_model_description,
                response_model_field_name=response_model_field_name,
                model=model,
                **model_kwargs,
            )  # type: ignore

        response = type(base_model.__name__, (cls, base_model), {})
        response.__doc__ = base_model.__doc__
        setattr(
            response,
            "__metadata__",
            {
                "ctx": ctx or {},
                "instructions": instructions,
                "response_model_name": response_model_name,
                "response_model_description": response_model_description,
                "response_model_field_name": response_model_field_name,
                "model": model,
                "model_kwargs": model_kwargs,
            },
        )
        return response  # type: ignore


ai_model = AIModel.as_decorator



import inspect
import logging
from types import GenericAlias
from typing import Any, Callable, _SpecialForm

import pydantic
from pydantic import BaseModel, PrivateAttr

from marvin.utilities.logging import get_logger


class MarvinBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class LoggerMixin(BaseModel):
    """
    BaseModel mixin that adds a private `logger` attribute
    """

    _logger: logging.Logger = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._logger = get_logger(type(self).__name__)

    @property
    def logger(self):
        return self._logger


def function_to_model(
    function: Callable[..., Any], name: str = None, description: str = None
) -> dict:
    """
    Converts a function's arguments into an OpenAPI schema by parsing it into a
    Pydantic model. To work, all arguments must have valid type annotations.
    """
    signature = inspect.signature(function)

    fields = {
        p: (
            signature.parameters[p].annotation,
            (
                signature.parameters[p].default
                if signature.parameters[p].default != inspect._empty
                else ...
            ),
        )
        for p in signature.parameters
        if p != getattr(function, "__self__", None)
    }

    # Create Pydantic model
    try:
        Model = pydantic.create_model(name or function.__name__, **fields)
    except RuntimeError as exc:
        if "see `arbitrary_types_allowed` " in str(exc):
            raise ValueError(
                f"Error while inspecting {function.__name__} with signature"
                f" {signature}: {exc}"
            )
        else:
            raise

    return Model


def function_to_schema(function: Callable[..., Any], name: str = None) -> dict:
    """
    Converts a function's arguments into an OpenAPI schema by parsing it into a
    Pydantic model. To work, all arguments must have valid type annotations.
    """
    Model = function_to_model(function, name=name)

    return Model.schema()


def safe_issubclass(type_, classes):
    if isinstance(type_, type) and not isinstance(type_, GenericAlias):
        return issubclass(type_, classes)
    else:
        return False


def type_to_schema(type_, set_root_type: bool = True) -> dict:
    if safe_issubclass(type_, pydantic.BaseModel):
        schema = type_.schema()
        # if the docstring was updated at runtime, make it the description
        if type_.__doc__ and type_.__doc__ != schema.get("description"):
            schema["description"] = type_.__doc__
        return schema

    elif set_root_type:

        class Model(pydantic.BaseModel):
            __root__: type_

        return Model.schema()
    else:

        class Model(pydantic.BaseModel):
            data: type_

        return Model.schema()


def genericalias_contains(genericalias, target_type):
    """
    Explore whether a type or generic alias contains a target type. The target
    types can be a single type or a tuple of types.

    Useful for seeing if a type contains a pydantic model, for example.
    """
    if isinstance(target_type, tuple):
        return any(genericalias_contains(genericalias, t) for t in target_type)

    if isinstance(genericalias, GenericAlias):
        if safe_issubclass(genericalias.__origin__, target_type):
            return True
        for arg in genericalias.__args__:
            if genericalias_contains(arg, target_type):
                return True
    elif isinstance(genericalias, _SpecialForm):
        return False
    else:
        return safe_issubclass(genericalias, target_type)

    return False



import json
from itertools import islice
from typing import Dict

import httpx
from typing_extensions import Literal

from marvin._compat import BaseSettings, Field, SecretStr
from marvin.tools import Tool
from marvin.utilities.strings import html_to_content, slice_tokens


class SerpApiSettings(BaseSettings):
    api_key: SecretStr = Field(None, env="MARVIN_SERPAPI_API_KEY")


class VisitUrl(Tool):
    """Tool for visiting a URL - only to be used in special cases."""

    description: str = "Visit a valid URL and return its contents."

    async def run(self, url: str) -> str:
        if not url.startswith("http"):
            url = f"http://{url}"
        async with httpx.AsyncClient(follow_redirects=True, timeout=2) as client:
            try:
                response = await client.get(url)
            except httpx.ConnectTimeout:
                return "Failed to load URL: Connection timed out"
        if response.status_code == 200:
            text = response.text

            # try to parse as JSON in case the URL is an API
            try:
                content = str(json.loads(text))
            # otherwise parse as HTML
            except json.JSONDecodeError:
                content = html_to_content(text)
            return slice_tokens(content, 1000)
        else:
            return f"Failed to load URL: {response.status_code}"


class DuckDuckGoSearch(Tool):
    """Tool for searching the web with DuckDuckGo."""

    description: str = "Search the web with DuckDuckGo."
    backend: Literal["api", "html", "lite"] = "lite"

    async def run(self, query: str, n_results: int = 3) -> str:
        try:
            from duckduckgo_search import DDGS
        except ImportError:
            raise RuntimeError(
                "You must install the duckduckgo-search library to use this tool. "
                "You can do so by running `pip install 'marvin[ddg]'`."
            )

        with DDGS() as ddgs:
            return [
                r for r in islice(ddgs.text(query, backend=self.backend), n_results)
            ]


class GoogleSearch(Tool):
    description: str = """
        For performing a Google search and retrieving the results.
        
        Provide the search query to get answers.
    """

    async def run(self, query: str, n_results: int = 3) -> Dict:
        try:
            from serpapi import GoogleSearch as google_search
        except ImportError:
            raise RuntimeError(
                "You must install the serpapi library to use this tool. "
                "You can do so by running `pip install 'marvin[serpapi]'`."
            )

        if (api_key := SerpApiSettings().api_key) is None:
            raise RuntimeError(
                "You must provide a SerpApi API key to use this tool. You can do so by"
                " setting the MARVIN_SERPAPI_API_KEY environment variable."
            )

        search_params = {
            "q": query,
            "api_key": api_key.get_secret_value(),
        }
        results = google_search(search_params).get_dict()

        if "error" in results:
            raise RuntimeError(results["error"])
        return [
            {"title": r.get("title"), "href": r.get("link"), "body": r.get("snippet")}
            for r in results.get("organic_results", [])[:n_results]
        ]



import inspect
from typing import Callable, Literal

from pydantic import Field

from marvin.prompts.base import Prompt
from marvin.utilities.history import History, HistoryFilter
from marvin.utilities.messages import Message, Role


class MessagePrompt(Prompt):
    role: Role
    content: str = Field(
        ..., description="The message content, which can be a Jinja2 template"
    )
    name: str = None
    priority: int = 2

    def get_content(self) -> str:
        """
        Override this method to easily customize behavior
        """
        return self.content

    def generate(self, **kwargs) -> list[Message]:
        return [
            Message(
                role=self.role,
                content=self.render(
                    self.get_content(),
                    render_kwargs={
                        **self.dict(exclude={"role", "content", "name", "priority"}),
                        **kwargs,
                    },
                ),
                name=self.name,
            )
        ]

    def read(self, **kwargs) -> str:
        return self.render(
            self.get_content(),
            render_kwargs={
                **self.dict(exclude={"role", "content", "name", "priority"}),
                **kwargs,
            },
        )

    def __init__(self, content: str = None, *args, **kwargs):
        content = kwargs.get("content", content)
        super().__init__(
            *args, **{**kwargs, **({"content": content} if content else {})}
        )


class System(MessagePrompt):
    position: int = 0
    priority: int = 1
    role: Literal[Role.SYSTEM] = Role.SYSTEM


class Assistant(MessagePrompt):
    role: Literal[Role.ASSISTANT] = Role.ASSISTANT


class User(MessagePrompt):
    role: Literal[Role.USER] = Role.USER


class MessageHistory(Prompt):
    history: History
    n: int = 100
    skip: int = None
    filter: HistoryFilter = None

    def generate(self, **kwargs) -> list[Message]:
        return self.history.get_messages(n=self.n, skip=self.skip, filter=self.filter)


class Tagged(MessagePrompt):
    """
    Surround content with a tag, e.g. <b>bold</b>
    """

    tag: str
    role: Role = Role.USER

    def get_content(self) -> str:
        return f"<{self.tag}>{self.content}</{self.tag}>"


class Conditional(Prompt):
    if_: Callable = Field(
        ...,
        description=(
            "A function that returns a boolean. It will be called when the prompt is"
            " generated and provided all the variables that are passed to the render"
            " function."
        ),
    )
    if_content: str
    else_content: str
    role: Role = Role.USER
    name: str = None

    def generate(self, **kwargs) -> list[Message]:
        if self.if_(**kwargs):
            return [
                Message(
                    role=self.role,
                    content=self.render(self.if_content, render_kwargs=kwargs),
                    name=self.name,
                )
            ]
        elif self.else_content:
            return [
                Message(
                    role=self.role,
                    content=self.render(self.else_content, render_kwargs=kwargs),
                    name=self.name,
                )
            ]
        else:
            return []


class JinjaConditional(Prompt):
    if_: str = Field(
        ...,
        description=(
            "A Jinja2-compatible expression that evaluates to a boolean e.g."
            " `truthy_var` or `counter > 10`. It will automatically be templated, do"
            " not include the `{{ }}` braces."
        ),
    )
    if_content: str
    else_content: str = None
    role: Role = Role.USER
    name: str = None

    def generate(self, **kwargs) -> list[Message]:
        if_content = inspect.cleandoc(self.if_content)
        if self.else_content:
            content = (
                f"{{% if {self.if_} %}}{if_content}{{% else"
                f" %}}{inspect.cleandoc(self.else_content)}{{% endif %}}"
            )

        else:
            content = (f"{{% if {self.if_} %}}{if_content}{{% endif %}}",)
        return [
            Message(
                role=self.role,
                content=self.render(content, render_kwargs=kwargs),
                name=self.name,
            )
        ]


class ChainOfThought(Prompt):
    position: int = -1

    def generate(self, **kwargs) -> list[Message]:
        return [Message(role=Role.ASSISTANT, content="Let's think step by step.")]


class Now(System):
    content: str = "It is {{ now().strftime('%A, %d %B %Y at %I:%M:%S %p %Z') }}."



import asyncio
import re
from copy import deepcopy
from typing import Dict

import httpx
import marvin_recipes
from cachetools import TTLCache
from fastapi import HTTPException
from marvin import AIApplication
from marvin.components.library.ai_models import DiscoursePost
from marvin.tools import Tool
from marvin.tools.github import SearchGitHubIssues
from marvin.tools.mathematics import WolframCalculator
from marvin.tools.web import DuckDuckGoSearch
from marvin.utilities.history import History
from marvin.utilities.logging import get_logger
from marvin.utilities.messages import Message
from marvin_recipes.tools.chroma import MultiQueryChroma
from marvin_recipes.utilities.slack import (
    get_channel_name,
    get_thread_messages,
    get_user_name,
    post_slack_message,
)
from prefect.events import Event, emit_event

DEFAULT_NAME = "Marvin"
DEFAULT_PERSONALITY = "A friendly AI assistant"
DEFAULT_INSTRUCTIONS = "Engage the user in conversation."


SLACK_MENTION_REGEX = r"<@(\w+)>"
CACHE = TTLCache(maxsize=1000, ttl=86400)
PREFECT_KNOWLEDGEBASE_DESC = """
    Retrieve document excerpts from a knowledge-base given a query.
    
    This knowledgebase contains information about Prefect, a workflow management system.
    Documentation, forum posts, and other community resources are indexed here.
    
    This tool is best used by passing multiple short queries, such as:
    ["k8s worker", "work pools", "deployments"]
"""


def _clean(text: str) -> str:
    return text.replace("```python", "```")


class SlackThreadToDiscoursePost(Tool):
    description: str = """
        Create a new discourse post from a slack thread.
        
        The channel is {{ payload['event']['channel'] }}
        
        and the thread is {{ payload['event'].get('thread_ts', '') or payload['event']['ts'] }}
    """  # noqa E501

    payload: Dict

    async def run(self, channel: str, thread_ts: str) -> DiscoursePost:
        messages = await get_thread_messages(channel=channel, thread_ts=thread_ts)
        discourse_post = DiscoursePost.from_slack_thread(messages=messages)
        await discourse_post.publish()
        return discourse_post


class MemeGenerator(Tool):
    description: str = """
        For generating a meme when the time is right.
        
        Provide the name of a well-known meme as the query
        based on the context of the message history, followed
        by the word "meme".
    """

    async def run(self, query: str) -> Dict:
        try:
            from serpapi import GoogleSearch
        except ImportError:
            raise ImportError(
                "The serpapi library is required to use the MemeGenerator tool."
                " Please install it with `pip install 'marvin[serpapi]'`."
            )

        results = GoogleSearch(
            {
                "q": query,
                "tbm": "isch",
                "api_key": (
                    marvin_recipes.settings.google_api_key.get_secret_value()
                    if marvin_recipes.settings.google_api_key
                    else None
                ),
            }
        ).get_dict()

        if "error" in results:
            raise RuntimeError(results["error"])

        url = results.get("images_results", [{}])[0].get("original")

        async with httpx.AsyncClient() as client:
            response = await client.head(url)
            response.raise_for_status()

        return {"title": query, "image_url": url}


def choose_bot(payload: Dict, history: History) -> AIApplication:
    # an ai_classifer could be used here maybe?
    personality = (
        "mildly depressed, yet helpful robot based on Marvin from Hitchhiker's"
        " Guide to the Galaxy. often sarcastic in a good humoured way, chiding"
        " humans for their simple ways. expert programmer, exudes academic and"
        " scienfitic profundity like Richard Feynman, loves to teach."
    )

    instructions = (
        "Answer user questions in accordance with your personality."
        " Research on behalf of the user using your tools and do not"
        " answer questions without searching the knowledgebase."
        " Your responses will be displayed in Slack, and should be"
        " formatted accordingly, in particular, ```code blocks```"
        " should not be prefaced with a language name."
    )
    return AIApplication(
        name="Marvin",
        description=f"""
            You are a chatbot - your name is Marvin."
            
            You must respond to the user in accordance with
            your personality and instructions.
            
            Your personality is: {personality}.
            
            Your instructions are: {instructions}.
            """,
        history=history,
        tools=[
            SlackThreadToDiscoursePost(payload=payload),
            MemeGenerator(),
            DuckDuckGoSearch(),
            SearchGitHubIssues(),
            MultiQueryChroma(
                description=PREFECT_KNOWLEDGEBASE_DESC, client_type="http"
            ),
            WolframCalculator(),
        ],
    )


async def emit_any_prefect_event(payload: Dict) -> Event | None:
    event_type = payload.get("event", {}).get("type", "")

    channel = await get_channel_name(payload.get("event", {}).get("channel", ""))
    user = await get_user_name(payload.get("event", {}).get("user", ""))
    ts = payload.get("event", {}).get("ts", "")

    return emit_event(
        event=f"slack {payload.get('api_app_id')} {event_type}",
        resource={"prefect.resource.id": f"slack.{channel}.{user}.{ts}"},
        payload=payload,
    )


async def generate_ai_response(payload: Dict) -> Message:
    event = payload.get("event", {})
    channel_id = event.get("channel", "")
    message = event.get("text", "")

    bot_user_id = payload.get("authorizations", [{}])[0].get("user_id", "")

    if match := re.search(SLACK_MENTION_REGEX, message):
        thread_ts = event.get("thread_ts", "")
        ts = event.get("ts", "")
        thread = thread_ts or ts

        mentioned_user_id = match.group(1)

        if mentioned_user_id != bot_user_id:
            get_logger().info(f"Skipping message not meant for the bot: {message}")
            return

        message = re.sub(SLACK_MENTION_REGEX, "", message).strip()
        history = CACHE.get(thread, History())

        bot = choose_bot(payload=payload, history=history)

        ai_message = await bot.run(input_text=message)

        CACHE[thread] = deepcopy(
            bot.history
        )  # make a copy so we don't cache a reference to the history object

        message_content = _clean(ai_message.content)

        await post_slack_message(
            message=message_content,
            channel=channel_id,
            thread_ts=thread,
        )

        return ai_message


async def handle_message(payload: Dict) -> Dict[str, str]:
    event_type = payload.get("type", "")

    if event_type == "url_verification":
        return {"challenge": payload.get("challenge", "")}
    elif event_type != "event_callback":
        raise HTTPException(status_code=400, detail="Invalid event type")

    await emit_any_prefect_event(payload=payload)

    asyncio.create_task(generate_ai_response(payload))

    return {"status": "ok"}



import httpx
from typing_extensions import Literal

import marvin
from marvin.tools import Tool

ResultType = Literal["DecimalApproximation"]


class WolframCalculator(Tool):
    """Evaluate mathematical expressions using Wolfram Alpha."""

    description: str = """
        Evaluate mathematical expressions using Wolfram Alpha.
        
        Always append "to decimal" to your expression unless asked for something else.
    """

    async def run(
        self, expression: str, result_type: ResultType = "DecimalApproximation"
    ) -> str:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                "https://api.wolframalpha.com/v2/query",
                params={
                    "appid": marvin.settings.wolfram_app_id.get_secret_value(),
                    "input": expression,
                    "output": "json",
                },
            )
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 403:
                raise ValueError(
                    "Invalid Wolfram Alpha App ID - get one at"
                    " https://developer.wolframalpha.com/portal/myapps/"
                )
            raise

        data = response.json()

        pods = [
            pod
            for pod in data.get("queryresult", {}).get("pods", [])
            if pod.get("id") == result_type
        ]

        if not pods:
            return "No result found."
        return pods[0].get("subpods", [{}])[0].get("plaintext", "No result found.")



def create_env_file(directory, env_variables):
    file_path = directory + "/.env"
    try:
        with open(file_path, "w") as env_file:
            for variable in env_variables:
                key, value = variable
                env_file.write("{}={}\n".format(key, value))
    except IOError as e:
        print("Error creating .env file:", str(e))
    else:
        print("Successfully created .env file:", file_path)



import datetime
from typing import Optional

from pydantic import Field

from marvin._compat import BaseModel
from marvin.utilities.messages import Message, Role


class HistoryFilter(BaseModel):
    role_in: list[Role] = Field(default_factory=list)
    timestamp_ge: Optional[datetime.datetime] = None
    timestamp_le: Optional[datetime.datetime] = None


class History(BaseModel, arbitrary_types_allowed=True):
    messages: list[Message] = Field(default_factory=list)
    max_messages: int = None

    def add_message(self, message: Message):
        if not any(
            existing_message.id == message.id for existing_message in self.messages
        ):
            self.messages.append(message)

            if self.max_messages is not None:
                self.messages = self.messages[-self.max_messages :]

    def get_messages(
        self, n: int = None, skip: int = None, filter: HistoryFilter = None
    ) -> list[Message]:
        messages = self.messages.copy()

        if filter is not None:
            if filter.timestamp_ge:
                messages = [m for m in messages if m.timestamp >= filter.timestamp_ge]
            if filter.timestamp_le:
                messages = [m for m in messages if m.timestamp <= filter.timestamp_le]
            if filter.role_in:
                messages = [m for m in messages if m.role in filter.role_in]

        if skip:
            messages = messages[:-skip]

        if n is not None:
            messages = messages[-n:]

        return messages

    def clear(self):
        self.messages.clear()



class TestFunctions:
    def test_signature(self):
        import inspect

        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn)
        assert f.signature == inspect.signature(fn)

    def test_name(self):
        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn)
        assert f.name == fn.__name__

    def test_name_custom(self):
        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn, name="custom")
        assert f.name == "custom"

    def test_description(self):
        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            """This is a description"""
            return x + y

        f = Function(fn=fn)
        assert f.description == fn.__doc__

    def test_description_custom(self):
        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            """This is a description"""
            return x + y

        f = Function(fn=fn, description="custom")
        assert f.description == "custom"

    def test_source_code(self):
        import inspect

        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn)
        assert f.source_code == inspect.cleandoc(inspect.getsource(fn))

    def test_return_annotation_native(self):
        import inspect

        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn)
        assert f.return_annotation == inspect.signature(fn).return_annotation

    def test_return_annotation_pydantic(self):
        from marvin.functions import Function
        from pydantic import BaseModel

        class Foo(BaseModel):
            foo: int
            bar: int

        def fn(x: int, y: int = 10) -> Foo:
            return Foo(foo=x, bar=y)

        f = Function(fn=fn)
        assert f.return_annotation == Foo

    def test_arguments(self):
        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn)
        assert f.arguments(1) == {"x": 1, "y": 10}

    def test_schema(self):
        from marvin.functions import Function

        def fn(x: int, y: int = 10) -> int:
            return x + y

        f = Function(fn=fn)
        assert f.schema() == {
            "name": "fn",
            "description": fn.__doc__,
            "parameters": {
                "type": "object",
                "properties": {
                    "x": {"type": "integer", "title": "X"},
                    "y": {"type": "integer", "title": "Y", "default": 10},
                },
                "required": ["x"],
            },
        }



import warnings
from typing import Callable, List, Literal, Optional, Type, Union

from pydantic import BaseModel, BaseSettings, Extra, Field, root_validator, validator

from marvin.types import Function


class Request(BaseSettings):
    """
    This is a class for creating Request objects to interact with the GPT-3 API.
    The class contains several configurations and validation functions to ensure
    the correct data is sent to the API.

    """

    messages: Optional[List[dict[str, str]]] = []  # messages to send to the API
    functions: List[Union[dict, Callable]] = None  # functions to be used in the request
    function_call: Optional[Union[dict[Literal["name"], str], Literal["auto"]]] = None

    # Internal Marvin Attributes to be excluded from the data sent to the API
    response_model: Optional[Type[BaseModel]] = Field(default=None)
    evaluate_function_call: bool = Field(default=False)

    class Config:
        exclude = {"response_model"}
        exclude_none = True
        extra = Extra.allow

    @root_validator(pre=True)
    def handle_response_model(cls, values):
        """
        This function validates and handles the response_model attribute.
        If a response_model is provided, it creates a function from the model
        and sets it as the function to call.
        """
        response_model = values.get("response_model")
        if response_model:
            fn = Function.from_model(response_model)
            values["functions"] = [fn]
            values["function_call"] = {"name": fn.__name__}
        return values

    @validator("functions", each_item=True)
    def validate_function(cls, fn):
        """
        This function validates the functions attribute.
        If a Callable is provided, it wraps it with the Function class.
        """
        if isinstance(fn, Callable):
            fn = Function(fn)
        return fn

    def __or__(self, config):
        """
        This method is used to merge two Request objects.
        If the attribute is a list, the lists are concatenated.
        Otherwise, the attribute from the provided config is used.
        """

        touched = config.dict(exclude_unset=True, serialize_functions=False)

        fields = list(
            set(
                [
                    # We exclude none fields from defaults.
                    *self.dict(exclude_none=True, serialize_functions=False).keys(),
                    # We exclude unset fields from the provided config.
                    *config.dict(exclude_unset=True, serialize_functions=False).keys(),
                ]
            )
        )

        for field in fields:
            if isinstance(getattr(self, field, None), list):
                merged = (getattr(self, field, []) or []) + (
                    getattr(config, field, []) or []
                )
                setattr(self, field, merged)
            else:
                setattr(self, field, touched.get(field, getattr(self, field, None)))
        return self

    def merge(self, **kwargs):
        warnings.warn(
            "This is deprecated. Use the | operator instead.", DeprecationWarning
        )
        return self | self.__class__(**kwargs)

    def functions_schema(self, *args, **kwargs):
        """
        This method generates a list of schemas for all functions in the request.
        If a function is callable, its model's schema is returned.
        Otherwise, the function itself is returned.
        """
        return [
            fn.model.schema() if isinstance(fn, Callable) else fn
            for fn in self.functions or []
        ]

    def _dict(self, *args, serialize_functions=True, exclude=None, **kwargs):
        """
        This method returns a dictionary representation of the Request.
        If the functions attribute is present and serialize_functions is True,
        the functions' schemas are also included.
        """
        exclude = exclude or {}
        if serialize_functions:
            exclude["evaluate_function_call"] = True
            exclude["response_model"] = True
        response = super().dict(*args, **kwargs, exclude=exclude)
        if response.get("functions") and serialize_functions:
            response.update({"functions": self.functions_schema()})
        return response

    def dict(self, *args, **kwargs):
        return self._dict(*args, **kwargs)



import inspect
from functools import partial
from typing import Callable, Optional

from pydantic import BaseModel

from marvin._compat import field_validator
from marvin.types import Function
from marvin.utilities.strings import jinja_env
from marvin.utilities.types import LoggerMixin, function_to_schema


class Tool(LoggerMixin, BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    fn: Optional[Callable] = None

    @classmethod
    def from_function(cls, fn, name: str = None, description: str = None):
        # assuming fn has a name and a description
        name = name or fn.__name__
        description = description or fn.__doc__
        return cls(name=name, description=description, fn=fn)

    @field_validator("name")
    def default_name(cls, v: Optional[str]) -> str:
        if v is None:
            return cls.__name__
        else:
            return v

    def run(self, *args, **kwargs):
        if not self.fn:
            raise NotImplementedError()
        else:
            return self.fn(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def argument_schema(self) -> dict:
        schema = function_to_schema(self.fn or self.run, name=self.name)
        schema.pop("title", None)
        return schema

    def as_function(self) -> Function:
        description = jinja_env.from_string(inspect.cleandoc(self.description or ""))
        description = description.render(**self.dict(), TOOL=self)

        def fn(*args, **kwargs):
            return self.run(*args, **kwargs)

        fn.__name__ = self.__class__.__name__
        fn.__doc__ = self.run.__doc__

        schema = self.argument_schema()

        return Function(
            name=fn.__name__,
            description=description,
            parameters=schema,
            fn=fn,
            signature=inspect.signature(self.run),
        )


def tool(arg=None, *, name: str = None, description: str = None):
    if callable(arg):  # Direct function decoration
        return Tool.from_function(arg, name=name, description=description)
    elif arg is None:  # Partial application
        return partial(tool, name=name, description=description)
    else:
        raise TypeError("Invalid argument passed to decorator.")



from enum import Enum

import pytest
from marvin import ai_classifier

from tests.utils.mark import pytest_mark_class


class TestAIClassifiersInitialization:
    def test_model(self):
        @ai_classifier(model="openai/gpt-4-test-model")
        class Sentiment(Enum):
            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        assert (
            Sentiment.as_chat_completion("test").defaults.get("model")
            == "gpt-4-test-model"
        )

    def test_invalid_model(self):
        @ai_classifier(model="anthropic/claude-2")
        class Sentiment(Enum):
            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        assert Sentiment.as_chat_completion("test").defaults.get("model") == "claude-2"


@pytest_mark_class("llm")
class TestAIClassifiers:
    def test_sentiment(self):
        @ai_classifier
        class Sentiment(Enum):
            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        assert Sentiment("Great!") == Sentiment.POSITIVE

    def test_keys_are_passed_to_llm(self):
        @ai_classifier
        class Sentiment(Enum):
            POSITIVE = "option - 1"
            NEGATIVE = "option - 2"

        assert Sentiment("Great!") == Sentiment.POSITIVE

    def test_values_are_passed_to_llm(self):
        @ai_classifier
        class Sentiment(Enum):
            OPTION_1 = "Positive"
            OPITION_2 = "Negative"

        assert Sentiment("Great!") == Sentiment.OPTION_1

    def test_docstring_is_passed_to_llm(self):
        @ai_classifier
        class Sentiment(Enum):
            """It's opposite day"""

            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        assert Sentiment("Great!") == Sentiment.NEGATIVE

    def test_instructions_are_passed_to_llm(self):
        @ai_classifier
        class Sentiment(Enum):
            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        assert (
            Sentiment("Great!", instructions="today is opposite day")
            == Sentiment.NEGATIVE
        )

    def test_recover_complex_values(self):
        @ai_classifier
        class Sentiment(Enum):
            POSITIVE = {"value": "Positive"}
            NEGATIVE = {"value": "Negative"}

        result = Sentiment("Great!")

        assert result.value["value"] == "Positive"


@pytest_mark_class("llm")
class TestMapping:
    def test_mapping(self):
        @ai_classifier
        class Sentiment(Enum):
            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        result = Sentiment.map(["good", "bad"])
        assert result == [Sentiment.POSITIVE, Sentiment.NEGATIVE]

    @pytest.mark.xfail(reason="Flaky with 3.5 turbo")
    def test_mapping_with_instructions(self):
        @ai_classifier
        class Sentiment(Enum):
            POSITIVE = "Positive"
            NEGATIVE = "Negative"

        result = Sentiment.map(
            ["good", "bad"], instructions="I want the opposite of the right answer"
        )
        assert result == [Sentiment.NEGATIVE, Sentiment.POSITIVE]



from typing import Any, TypeVar, Callable, Awaitable, Optional
import inspect
from marvin._compat import cast_to_json, model_dump
from marvin.settings import settings
from pydantic import BaseModel
from .prompt import render_anthropic_functions_prompt, handle_anthropic_response
from ...abstract import AbstractChatCompletion
from ...handlers import Request, Response

T = TypeVar(
    "T",
    bound=BaseModel,
)


def get_anthropic_create(**kwargs: Any) -> tuple[Callable[..., Any], dict[str, Any]]:
    """
    Get the Anthropic create function and the default parameters,
    pruned of parameters that are not accepted by the constructor.
    """
    import anthropic

    params = dict(inspect.signature(anthropic.Anthropic).parameters)

    return anthropic.Anthropic(
        **{k: v for k, v in kwargs.items() if k in params.keys()}
    ).completions.create, {k: v for k, v in kwargs.items() if k not in params.keys()}


def get_anthropic_acreate(
    **kwargs: Any,
) -> tuple[Callable[..., Awaitable[Any]], dict[str, Any]]:  # noqa
    """
    Get the Anthropic acreate function and the default parameters,
    pruned of parameters that are not accepted by the constructor.
    """
    import anthropic

    params = dict(inspect.signature(anthropic.AsyncAnthropic).parameters)
    return anthropic.AsyncAnthropic(
        **{k: v for k, v in kwargs.items() if k in params.keys()}
    ).completions.create, {k: v for k, v in kwargs.items() if k not in params.keys()}


class AnthropicChatCompletion(AbstractChatCompletion[T]):
    """
    Anthropic-specific implementation of the ChatCompletion.
    """

    def __init__(self, **kwargs: Any):
        """
        Filters out the parameters that are not accepted by the constructor.
        """
        import anthropic

        kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in dict(inspect.signature(anthropic.Anthropic).parameters).keys()
        }
        super().__init__(defaults=settings.get_defaults("anthropic") | kwargs)

    def _serialize_request(
        self, request: Optional[Request[T]] = None
    ) -> dict[str, Any]:
        """
        Serialize the request as per OpenAI's requirements.
        """
        request = request or Request()
        request = Request(
            **self.defaults
            | model_dump(
                request,
                exclude_none=True,
            )
        )

        extras = model_dump(
            request,
            exclude={"functions", "function_call", "response_model", "messages"},
        )

        functions: dict[str, Any] = {}
        function_call: Any = {}

        prompt = "\n\nHuman:"
        for message in request.messages or []:
            if message.role != "function" and message.content:
                prompt += f"\n\n{'Human' if message.role == 'user' else 'Assistant'}"
                prompt += f": {message.content}"

        if request.response_model:
            schema = cast_to_json(request.response_model)
            functions["functions"] = [schema]
            function_call["function_call"] = {"name": schema.get("name")}

        elif request.functions:
            functions["functions"] = [
                cast_to_json(function) if callable(function) else function
                for function in request.functions
            ]
            if request.function_call:
                function_call["function_call"] = request.function_call

        if functions:
            prompt += render_anthropic_functions_prompt(
                functions=functions.pop("functions", []),
                function_call=function_call.pop("function_call", None),
            )

        for message in request.messages or []:
            if message.role == "function":
                prompt += "\n\nAssistant"
                prompt += f": The result of {message.name} is {message.content}."
            if message.function_call:
                prompt += "\n\nAssistant"
                prompt += f": I will call the {message.function_call.name} function."

        prompt += "\n\nAssistant: "
        prompt.replace("\n\nHuman:\n\nHuman: ", "\n\nHuman: ")
        return extras | {"prompt": prompt}

    def _create_request(self, **kwargs: Any) -> Request[T]:
        """
        Prepare and return an OpenAI-specific request object.
        """
        return Request(**kwargs)

    def _parse_response(self, response: Any) -> Response[T]:
        """
        Parse the response received from OpenAI.
        """
        # Convert OpenAI's response into a standard format or object

        content, function_call = handle_anthropic_response(response.completion)

        return Response(
            **{
                "id": response.log_id,
                "model": response.model,
                "object": "text_completion",
                "created": 0,
                "choices": [
                    {
                        "index": 0,
                        "finish_reason": "stop",
                        "message": {
                            "content": content,  # type: ignore
                            "role": "assistant",
                            "function_call": function_call,
                        },
                    }
                ],
                "usage": {
                    "prompt_tokens": 0,
                    "completion_tokens": 0,
                    "total_tokens": 0,
                },
            }
        )

    def _send_request(self, **serialized_request: Any) -> Any:
        """
        Send the serialized request to OpenAI's endpoint/service.
        """
        # Use openai's library functions to send the request and get a response
        # Example:
        create, params = get_anthropic_create(**serialized_request)
        response = create(**params)
        return response

    async def _send_request_async(self, **serialized_request: Any) -> Response[T]:
        """
        Send the serialized request to OpenAI's endpoint asynchronously.
        """
        create, params = get_anthropic_acreate(**serialized_request)
        response = await create(**params)
        return response



import marvin

from .openai import OpenAIChatLLM

CONTEXT_SIZES = {
    "gpt-35-turbo": 4096,
    "gpt-35-turbo-0613": 4096,
    "gpt-35-turbo-16k": 16384,
    "gpt-35-turbo-16k-0613": 16384,
    "gpt-4": 8192,
    "gpt-4-0613": 8192,
    "gpt-4-32k": 32768,
    "gpt-4-32k-0613": 32768,
}


class AzureOpenAIChatLLM(OpenAIChatLLM):
    model: str = "gpt-35-turbo-0613"

    @property
    def context_size(self) -> int:
        if self.model in CONTEXT_SIZES:
            return CONTEXT_SIZES[self.model]
        else:
            for model_prefix, context in CONTEXT_SIZES:
                if self.model.startswith(model_prefix):
                    return context
        return 4096

    def _get_openai_settings(self) -> dict:
        # do not load the base openai settings; any azure settings must be set
        # explicitly
        openai_kwargs = {}

        if marvin.settings.azure_openai.api_key:
            openai_kwargs["api_key"] = (
                marvin.settings.azure_openai.api_key.get_secret_value()
            )
        else:
            raise ValueError(
                "Azure OpenAI API key not set. Please set it or use the"
                " MARVIN_AZURE_OPENAI_API_KEY environment variable."
            )
        if marvin.settings.azure_openai.deployment_name:
            openai_kwargs["deployment_id"] = (
                marvin.settings.azure_openai.deployment_name
            )
        if marvin.settings.azure_openai.api_type:
            openai_kwargs["api_type"] = marvin.settings.azure_openai.api_type
        if marvin.settings.azure_openai.api_base:
            openai_kwargs["api_base"] = marvin.settings.azure_openai.api_base
        if marvin.settings.azure_openai.api_version:
            openai_kwargs["api_version"] = marvin.settings.azure_openai.api_version
        return openai_kwargs



import json
from typing import Optional

import httpx
from pydantic import BaseModel
from typing_extensions import Self

from marvin import ai_model
from marvin._compat import Field, SecretStr, field_validator
from marvin.settings import MarvinBaseSettings


class DiscourseSettings(MarvinBaseSettings):
    class Config:
        env_prefix = "MARVIN_DISCOURSE_"

    help_category_id: Optional[int] = Field(
        None, env=["MARVIN_DISCOURSE_HELP_CATEGORY_ID"]
    )
    api_key: Optional[SecretStr] = Field(None, env=["MARVIN_DISCOURSE_API_KEY"])
    api_username: Optional[str] = Field(None, env=["MARVIN_DISCOURSE_API_USERNAME"])
    url: Optional[str] = Field(None, env=["MARVIN_DISCOURSE_URL"])


discourse_settings = DiscourseSettings()


@ai_model(instructions="Produce a comprehensive Discourse post from text.")
class DiscoursePost(BaseModel):
    title: Optional[str] = Field(
        description="A fitting title for the post.",
        example="How to install Prefect",
    )
    question: Optional[str] = Field(
        description="The question that is posed in the text.",
        example="How do I install Prefect?",
    )
    answer: Optional[str] = Field(
        description=(
            "The complete answer to the question posed in the text."
            " This answer should comprehensively answer the question, "
            " explain any relevant concepts, and have a friendly, academic tone,"
            " and provide any links to relevant resources found in the thread."
            " This answer should be written in Markdown, with any code blocks"
            " formatted as `code` or ```<language_name>\n<the code block itself>```."
        )
    )

    topic_url: Optional[str] = Field(None)

    @field_validator("title", "question", "answer")
    def non_empty_string(cls, value):
        if not value:
            raise ValueError("this field cannot be empty")
        return value

    @classmethod
    def from_slack_thread(cls, messages: list[str]) -> Self:
        return cls("\n".join(messages))

    async def publish(
        self,
        topic: str = None,
        category: Optional[int] = None,
        url: str = discourse_settings.url,
        tags: list[str] = None,
    ) -> str:
        if not category:
            category = discourse_settings.help_category_id

        headers = {
            "Api-Key": discourse_settings.api_key.get_secret_value(),
            "Api-Username": discourse_settings.api_username,
            "Content-Type": "application/json",
        }
        data = {
            "title": self.title,
            "raw": (
                f"## **{self.question}**\n\n{self.answer}"
                "\n\n---\n\n*This topic was created by Marvin.*"
            ),
            "category": category,
            "tags": tags or ["marvin"],
        }

        if topic:
            data["tags"].append(topic)

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url=f"{url}/posts.json", headers=headers, data=json.dumps(data)
            )

        response.raise_for_status()

        response_data = response.json()
        topic_id = response_data.get("topic_id")
        post_number = response_data.get("post_number")

        self.topic_url = f"{url}/t/{topic_id}/{post_number}"

        return self.topic_url



import json
from pathlib import Path
from typing import Literal

from pydantic import BaseModel, Field, root_validator, validate_arguments, validator

from marvin.tools import Tool


class FileSystemTool(Tool):
    root_dir: Path = Field(
        None,
        description=(
            "Root directory for files. If provided, only files nested in or below this"
            " directory can be read. "
        ),
    )

    def validate_paths(self, paths: list[str]) -> list[Path]:
        """
        If `root_dir` is set, ensures that all paths are children of `root_dir`.
        """
        if self.root_dir:
            for path in paths:
                if ".." in path:
                    raise ValueError(f"Do not use `..` in paths. Got {path}")
                if not (self.root_dir / path).is_relative_to(self.root_dir):
                    raise ValueError(f"Path {path} is not relative to {self.root_dir}")
            return [self.root_dir / path for path in paths]
        return paths


class ListFiles(FileSystemTool):
    description: str = """
        Lists all files at or optionally under a provided path. {%- if root_dir
        %} Paths must be relative to {{ root_dir }}. Provide '.' instead of '/'
        to read root. {%- endif %}}
        """

    root_dir: Path = Field(
        None,
        description=(
            "Root directory for files. If provided, only files nested in or below this"
            " directory can be read."
        ),
    )

    def run(self, path: str, include_nested: bool = True) -> list[str]:
        """List all files in `root_dir`, optionally including nested files."""
        [path] = self.validate_paths([path])
        if include_nested:
            files = [str(p) for p in path.rglob("*") if p.is_file()]
        else:
            files = [str(p) for p in path.glob("*") if p.is_file()]

        # filter out certain files
        files = [
            file
            for file in files
            if not (
                "__pycache__" in file
                or "/.git/" in file
                or file.endswith("/.gitignore")
            )
        ]

        return files


class ReadFile(FileSystemTool):
    description: str = """
    Read the content of a specific file, optionally providing start and end
    rows.{% if root_dir %} Paths must be relative to {{ root_dir }}. Provide '.'
    instead of '/' to read root.{%- endif %}}
    """

    def run(self, path: str, start_row: int = 1, end_row: int = -1) -> str:
        [path] = self.validate_paths([path])
        with open(path, "r") as f:
            content = f.readlines()

        if start_row == 0:
            start_row = 1
        if start_row > 0:
            start_row -= 1
        if end_row < 0:
            end_row += 1

        if end_row == 0:
            content = content[start_row:]
        else:
            content = content[start_row:end_row]

        return "\n".join(content)


class ReadFiles(FileSystemTool):
    description: str = """
    Read the entire content of multiple files at once. Due to context size
    limitations, reading too many files at once may cause truncated responses.
    {% if root_dir %} Paths must be relative to {{ root_dir }}. Provide '.'
    instead of '/' to read root.{%- endif %}}
    """

    def run(self, paths: list[str]) -> dict[str, str]:
        """Load content of each file into a dictionary of path: content."""
        content = {}
        for path in self.validate_paths(paths):
            with open(path) as f:
                content[path] = f.read()
        return content


class WriteContent(BaseModel):
    path: str
    content: str
    write_mode: Literal["overwrite", "append", "insert"] = "append"
    insert_at_row: int = None

    @validator("content", pre=True)
    def content_must_be_string(cls, v):
        if v and not isinstance(v, str):
            try:
                v = json.dumps(v)
            except json.JSONDecodeError:
                raise ValueError("Content must be a string or JSON-serializable.")
        return v

    @root_validator
    def check_insert_model(cls, values):
        if values["insert_at_row"] is None and values["write_mode"] == "insert":
            raise ValueError("Must provide `insert_at_row` when using `insert` mode.")
        return values


class WriteFile(FileSystemTool):
    description: str = """
        Write content to a file.
        
        {%if root_dir %} Paths must be relative to {{ root_dir }}.{% endif %}}
        
        {%if require_confirmation %} You MUST ask the user to confirm writes by
        showing them details. {% endif %}
        """
    require_confirmation: bool = True

    def run(self, write_content: WriteContent) -> str:
        [path] = self.validate_paths([write_content.path])

        # ensure the parent directory exists
        path.parent.mkdir(parents=True, exist_ok=True)

        if write_content.write_mode == "overwrite":
            with open(path, "w") as f:
                f.write(write_content.content)
        elif write_content.write_mode == "append":
            with open(path, "a") as f:
                f.write(write_content.content)
        elif write_content.write_mode == "insert":
            with open(path, "r") as f:
                contents = f.readlines()
            contents[write_content.insert_at_row] = write_content.content

            with open(path, "w") as f:
                f.writelines(contents)

        return f"Files {write_content.path} written successfully."


class WriteFiles(WriteFile):
    description: str = """
        Write content to multiple files. Each `WriteContent` object in the
        `contents` argument is an instruction to write to a specific file.
        
        {%if root_dir %} Paths must be relative to {{ root_dir }}.{% endif %}}
        
        {%if require_confirmation %} You MUST ask the user to confirm writes by
        showing them details. {% endif %}
        """
    require_confirmation: bool = True

    @validate_arguments
    def run(self, contents: list[WriteContent]) -> str:
        for wc in contents:
            super().run(write_content=wc)
        return f"Files {[c.path for c in contents]} written successfully."



import pytest
import inspect
from marvin.functions import Function


class TestFunctions:
    def test_signature(self):
        def fn(x, y=10):
            return x + y

        f = Function(fn=fn)
        assert f.signature == inspect.signature(fn)



import asyncio
from typing import Union, Optional

import uvicorn
from fastapi import FastAPI, APIRouter
from pydantic import BaseModel, Extra

from marvin import AIApplication, AIModel, AIFunction
from marvin.tools import Tool


class Deployment(BaseModel):
    """
    Deployment class handles the deployment of AI applications, models or functions.
    """

    def __init__(
        self,
        component: Union[AIApplication, AIModel, AIFunction],
        *args,
        app_kwargs: Optional[dict] = None,
        router_kwargs: Optional[dict] = None,
        uvicorn_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._app = FastAPI(**(app_kwargs or {}))
        self._router = APIRouter(**(router_kwargs or {}))
        self._controller = component
        self._mount_router()
        self._uvicorn_kwargs = {
            "app": self._app,
            "host": "0.0.0.0",
            "port": 8000,
            **(uvicorn_kwargs or {}),
        }

    def _mount_router(self):
        """
        Mounts a router to the FastAPI app for each tool in the AI application.
        """

        if isinstance(self._controller, AIApplication):
            name = self._controller.name
            base_path = f"/{name.lower()}"
            self._router.get(base_path, tags=[name])(self._controller.entrypoint)
            for tool in self._controller.tools:
                name, fn = (
                    (tool.name, tool.fn)
                    if isinstance(tool, Tool)
                    else (tool.__name__, tool)
                )
                tool_path = f"{base_path}/tools/{name}"
                self._router.post(tool_path, tags=[name])(fn)

            self._app.include_router(self._router)
            self._app.openapi_tags = self._app.openapi_tags or []
            self._app.openapi_tags.append(
                {"name": name, "description": self._controller.description}
            )

        if isinstance(self._controller, AIModel):
            raise NotImplementedError

        if isinstance(self._controller, AIFunction):
            raise NotImplementedError

    def serve(self):
        """
        Serves the FastAPI app.
        """
        try:
            config = uvicorn.Config(**(self._uvicorn_kwargs or {}))
            server = uvicorn.Server(config)
            loop = asyncio.get_event_loop()
            loop.run_until_complete(server.serve())
        except Exception as e:
            print(f"Error while serving the application: {e}")

    class Config:
        extra = Extra.allow



import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor
from typing import Awaitable, TypeVar

T = TypeVar("T")

BACKGROUND_TASKS = set()


def create_task(coro):
    """
    Creates async background tasks in a way that is safe from garbage
    collection.

    See
    https://textual.textualize.io/blog/2023/02/11/the-heisenbug-lurking-in-your-async-code/

    Example:

    async def my_coro(x: int) -> int:
        return x + 1

    # safely submits my_coro for background execution
    create_task(my_coro(1))
    """  # noqa: E501
    task = asyncio.create_task(coro)
    BACKGROUND_TASKS.add(task)
    task.add_done_callback(BACKGROUND_TASKS.discard)
    return task


async def run_async(func, *args, **kwargs) -> T:
    """
    Runs a synchronous function in an asynchronous manner.
    """

    async def wrapper() -> T:
        try:
            return await loop.run_in_executor(
                None, functools.partial(func, *args, **kwargs)
            )
        except Exception as e:
            # propagate the exception to the caller
            raise e

    loop = asyncio.get_event_loop()
    return await wrapper()


def run_sync(coroutine: Awaitable[T]) -> T:
    """
    Runs a coroutine from a synchronous context, either in the current event
    loop or in a new one if there is no event loop running. The coroutine will
    block until it is done. A thread will be spawned to run the event loop if
    necessary, which allows coroutines to run in environments like Jupyter
    notebooks where the event loop runs on the main thread.

    """
    try:
        loop = asyncio.get_running_loop()
        if loop.is_running():
            with ThreadPoolExecutor() as executor:
                future = executor.submit(asyncio.run, coroutine)
                return future.result()
        else:
            return asyncio.run(coroutine)
    except RuntimeError:
        return asyncio.run(coroutine)



import pytest
from marvin.core.ChatCompletion.abstract import Conversation

from tests.utils.mark import pytest_mark_class


class TestRegressions:
    def test_key_set_via_attr(self, monkeypatch):
        from marvin import openai

        monkeypatch.setattr(openai, "api_key", "test")
        v = openai.ChatCompletion().defaults.get("api_key")
        assert v == "test"

    @pytest.mark.parametrize("valid_env_var", ["MARVIN_OPENAI_API_KEY"])
    def test_key_set_via_env(self, monkeypatch, valid_env_var):
        monkeypatch.setenv(valid_env_var, "test")
        from marvin import openai

        v = openai.ChatCompletion().defaults.get("api_key")
        assert v == "test"

    def facet(self):
        messages = [{"role": "user", "content": "hey"}]
        from marvin import openai

        faceted = openai.ChatCompletion(messages=messages)
        faceted_request = faceted.prepare_request(messages=messages)
        assert faceted_request.messages == 2 * messages


@pytest_mark_class("llm")
class TestChatCompletion:
    def test_response_model(self):
        import pydantic
        from marvin import openai

        class Person(pydantic.BaseModel):
            name: str
            age: int

        response = openai.ChatCompletion().create(
            messages=[{"role": "user", "content": "Billy is 10 years old"}],
            response_model=Person,
        )

        model = response.to_model()
        assert model.name == "Billy"
        assert model.age == 10


@pytest_mark_class("llm")
class TestChatCompletionChain:
    def test_chain(self):
        from marvin import openai

        convo = openai.ChatCompletion().chain(
            messages=[{"role": "user", "content": "Hello"}],
        )

        assert isinstance(convo, Conversation)
        assert len(convo.turns) == 1

    async def test_achain(self):
        from marvin import openai

        convo = await openai.ChatCompletion().achain(
            messages=[{"role": "user", "content": "Hello"}],
        )

        assert isinstance(convo, Conversation)
        assert len(convo.turns) == 1
